import dataclasses
import pathlib
import textwrap
import warnings

import yuio.color
import yuio.term
import yuio.theme

import pytest


class TestColors:
    def test_colors_immutable(self, theme):
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            theme.colors = {}
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            del theme.colors
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            theme.colors["foo"] = "bar"
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            del theme.colors["foo"]
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            theme.__class__.colors = {}
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            del theme.__class__.colors
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            theme.__class__.colors["foo"] = "bar"
        with pytest.raises(TypeError, match=r"Theme\.colors is immutable"):
            del theme.__class__.colors["foo"]

    def test_get_color(self, theme):
        assert theme.get_color("code") == yuio.color.Color.FORE_MAGENTA
        assert theme.get_color("note") == yuio.color.Color.FORE_CYAN
        assert theme.get_color("bold") == yuio.color.Color.STYLE_BOLD
        assert theme.get_color("b") == yuio.color.Color.STYLE_BOLD
        assert theme.get_color("dim") == yuio.color.Color.STYLE_DIM
        assert theme.get_color("d") == yuio.color.Color.STYLE_DIM
        assert theme.get_color("normal") == yuio.color.Color.FORE_NORMAL
        assert theme.get_color("red") == yuio.color.Color.FORE_RED
        assert theme.get_color("green") == yuio.color.Color.FORE_GREEN
        assert theme.get_color("yellow") == yuio.color.Color.FORE_YELLOW
        assert theme.get_color("blue") == yuio.color.Color.FORE_BLUE
        assert theme.get_color("magenta") == yuio.color.Color.FORE_MAGENTA
        assert theme.get_color("cyan") == yuio.color.Color.FORE_CYAN
        assert theme.get_color("#ff0000") == yuio.color.Color(
            fore=yuio.color.ColorValue.from_rgb(0xFF, 0x00, 0x00)
        )
        assert theme.get_color("bg#ff0000") == yuio.color.Color(
            back=yuio.color.ColorValue.from_rgb(0xFF, 0x00, 0x00)
        )
        assert theme.get_color("#00ff00 bg#ff0000") == yuio.color.Color(
            fore=yuio.color.ColorValue.from_rgb(0x00, 0xFF, 0x00),
            back=yuio.color.ColorValue.from_rgb(0xFF, 0x00, 0x00),
        )
        with pytest.warns(yuio.theme.ThemeWarning, match=r"#ffxxyy"):
            assert theme.get_color("#ffxxyy") == yuio.color.Color.NONE
        with pytest.warns(yuio.theme.ThemeWarning, match=r"bg#ffxxyy"):
            assert theme.get_color("bg#ffxxyy") == yuio.color.Color.NONE

    def test_to_color(self, theme):
        assert theme.to_color(None) == yuio.color.Color.NONE
        assert theme.to_color("blue") == yuio.color.Color.FORE_BLUE
        assert (
            theme.to_color(yuio.color.Color.FORE_MAGENTA)
            == yuio.color.Color.FORE_MAGENTA
        )

    def test_color_paths(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.color.Color.STYLE_BOLD,
                "x/y": yuio.color.Color.FORE_RED,
                "x/z": yuio.color.Color.FORE_GREEN,
                "y/y": yuio.color.Color.FORE_BLUE,
            }

        a = A()

        assert a.get_color("x") == yuio.color.Color.STYLE_BOLD
        assert (
            a.get_color("x/y")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_RED
        )
        assert (
            a.get_color("x/z")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_GREEN
        )
        assert a.get_color("y/y") == yuio.color.Color.FORE_BLUE

    def test_color_redirect(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_r": yuio.color.Color.FORE_RED,
                "t_g": yuio.color.Color.FORE_GREEN,
                "t_b": yuio.color.Color.STYLE_BOLD,
                "t_d": yuio.color.Color.STYLE_DIM,
                "b": yuio.color.Color.BACK_CYAN,
                "x": "t_b",
                "x/y": "t_r",
                "y": "t_d",
                "y/y": "x/y",
                "z": "x/y y/y b",
            }

        a = A()

        assert a.get_color("x") == yuio.color.Color.STYLE_BOLD
        assert (
            a.get_color("x/y")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_RED
        )
        assert a.get_color("y") == yuio.color.Color.STYLE_DIM
        assert (
            a.get_color("y/y")
            == yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.FORE_RED
        )
        assert (
            a.get_color("z")
            == yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.FORE_RED
            | yuio.color.Color.BACK_CYAN
        )

    def test_color_ctx(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.color.Color.STYLE_BOLD,
                "x:y": yuio.color.Color.FORE_RED,
                "x:y/z": yuio.color.Color.STYLE_DIM,
                "x:a": yuio.color.Color.FORE_GREEN,
                "q/q": yuio.color.Color.FORE_BLUE,
            }

        a = A()

        assert a.get_color("x") == yuio.color.Color.STYLE_BOLD
        assert (
            a.get_color("x:y")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_RED
        )
        assert (
            a.get_color("x:a")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_GREEN
        )
        assert (
            a.get_color("x/a/q:y")
            == yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_RED
        )
        assert (
            a.get_color("x:y/z")
            == yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_DIM
        )
        assert a.get_color("q/q") == yuio.color.Color.FORE_BLUE
        assert a.get_color("q/q:y") == yuio.color.Color.FORE_BLUE

    def test_simple_inheritance(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_a": yuio.color.Color.FORE_RED,
                "t_ab": yuio.color.Color.FORE_GREEN,
                "t_abc": yuio.color.Color.FORE_BLUE,
            }

        class B(yuio.theme.Theme):
            colors = {
                "t_b": yuio.color.Color.FORE_CYAN,
                "t_ab": yuio.color.Color.FORE_MAGENTA,
                "t_abc": yuio.color.Color.FORE_YELLOW,
            }

        class C(A, B):
            colors = {
                "t_c": yuio.color.Color.FORE_BLACK,
                "t_abc": yuio.color.Color.FORE_WHITE,
            }

        c = C()

        assert C.colors["t_a"] == yuio.color.Color.FORE_RED
        assert C.colors["t_b"] == yuio.color.Color.FORE_CYAN
        assert C.colors["t_c"] == yuio.color.Color.FORE_BLACK
        assert C.colors["t_ab"] == yuio.color.Color.FORE_GREEN
        assert C.colors["t_abc"] == yuio.color.Color.FORE_WHITE

        assert c.get_color("t_a") == yuio.color.Color.FORE_RED
        assert c.get_color("t_b") == yuio.color.Color.FORE_CYAN
        assert c.get_color("t_c") == yuio.color.Color.FORE_BLACK
        assert c.get_color("t_ab") == yuio.color.Color.FORE_GREEN
        assert c.get_color("t_abc") == yuio.color.Color.FORE_WHITE

    def test_color_overrides(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_a": yuio.color.Color.FORE_RED,
                "t_b": yuio.color.Color.FORE_GREEN,
                "t_b_2": yuio.color.Color.FORE_MAGENTA,
            }

            def __init__(self):
                super().__init__()

                self._set_color_if_not_overridden("t_a", yuio.color.Color.BACK_RED)
                self._set_color_if_not_overridden("t_b", yuio.color.Color.BACK_GREEN)
                self.set_color("t_b_2", yuio.color.Color.BACK_MAGENTA)

        class B(A):
            colors = {
                "t_b": yuio.color.Color.FORE_CYAN,
                "t_b_2": yuio.color.Color.FORE_CYAN,
            }

        class C(B):
            pass

        a = A()
        b = B()
        c = C()

        assert A.colors["t_a"] == yuio.color.Color.FORE_RED
        assert A.colors["t_b"] == yuio.color.Color.FORE_GREEN
        assert A.colors["t_b_2"] == yuio.color.Color.FORE_MAGENTA

        assert a.get_color("t_a") == yuio.color.Color.BACK_RED
        assert a.get_color("t_b") == yuio.color.Color.BACK_GREEN
        assert a.get_color("t_b_2") == yuio.color.Color.BACK_MAGENTA

        assert B.colors["t_a"] == yuio.color.Color.FORE_RED
        assert B.colors["t_b"] == yuio.color.Color.FORE_CYAN
        assert B.colors["t_b_2"] == yuio.color.Color.FORE_CYAN

        assert b.get_color("t_a") == yuio.color.Color.BACK_RED
        assert b.get_color("t_b") == yuio.color.Color.FORE_CYAN
        assert b.get_color("t_b_2") == yuio.color.Color.BACK_MAGENTA

        assert C.colors["t_a"] == yuio.color.Color.FORE_RED
        assert C.colors["t_b"] == yuio.color.Color.FORE_CYAN
        assert C.colors["t_b_2"] == yuio.color.Color.FORE_CYAN

        assert c.get_color("t_a") == yuio.color.Color.BACK_RED
        assert c.get_color("t_b") == yuio.color.Color.FORE_CYAN
        assert c.get_color("t_b_2") == yuio.color.Color.BACK_MAGENTA

    def test_set_color(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.color.Color.STYLE_BOLD,
            }

        a = A()

        assert a.get_color("x") == yuio.color.Color.STYLE_BOLD
        assert a.get_color("x/y") == yuio.color.Color.STYLE_BOLD
        assert a.get_color("z") == yuio.color.Color.NONE

        a.set_color("x", yuio.color.Color.FORE_RED)
        a.set_color("z", yuio.color.Color.FORE_BLUE)

        assert a.get_color("x") == yuio.color.Color.FORE_RED
        assert a.get_color("x/y") == yuio.color.Color.FORE_RED
        assert a.get_color("z") == yuio.color.Color.FORE_BLUE

    def test_set_color_if_not_overridden_error_outside_of_init(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.color.Color.STYLE_BOLD,
            }

        a = A()

        with pytest.raises(TypeError):
            a._set_color_if_not_overridden("x", yuio.color.Color.FORE_RED)


class TestDecorations:
    def test_unicode_decorations_immutable(self, theme):
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            theme.msg_decorations_unicode = {}
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            del theme.msg_decorations_unicode
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            theme.msg_decorations_unicode["foo"] = "bar"
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            del theme.msg_decorations_unicode["foo"]
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            theme.__class__.msg_decorations_unicode = {}
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            del theme.__class__.msg_decorations_unicode
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            theme.__class__.msg_decorations_unicode["foo"] = "bar"
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_unicode is immutable"
        ):
            del theme.__class__.msg_decorations_unicode["foo"]

    def test_ascii_decorations_immutable(self, theme):
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            theme.msg_decorations_ascii = {}
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            del theme.msg_decorations_ascii
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            theme.msg_decorations_ascii["foo"] = "bar"
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            del theme.msg_decorations_ascii["foo"]
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            theme.__class__.msg_decorations_ascii = {}
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            del theme.__class__.msg_decorations_ascii
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            theme.__class__.msg_decorations_ascii["foo"] = "bar"
        with pytest.raises(
            TypeError, match=r"Theme\.msg_decorations_ascii is immutable"
        ):
            del theme.__class__.msg_decorations_ascii["foo"]

    def test_get_msg_decoration(self, theme):
        assert theme.get_msg_decoration("heading/1", is_unicode=True) == "⣿ "
        assert theme.get_msg_decoration("heading/1", is_unicode=False) == "# "
        assert theme.get_msg_decoration("unknown", is_unicode=True) == ""
        assert theme.get_msg_decoration("unknown", is_unicode=False) == ""

    def test_decoration_overrides(self):
        class A(yuio.theme.Theme):
            msg_decorations_unicode = {
                "t_a": "a_u",
                "t_b": "a_u",
                "t_b_2": "a_u",
            }

            msg_decorations_ascii = {
                "t_a": "a_a",
                "t_b": "a_a",
                "t_b_2": "a_a",
            }

            def __init__(self):
                super().__init__()

                self._set_msg_decoration_unicode_if_not_overridden("t_a", "a2_u")
                self._set_msg_decoration_unicode_if_not_overridden("t_b", "a2_u")
                self.set_msg_decoration_unicode("t_b_2", "a2_u")

                self._set_msg_decoration_ascii_if_not_overridden("t_a", "a2_a")
                self._set_msg_decoration_ascii_if_not_overridden("t_b", "a2_a")
                self.set_msg_decoration_ascii("t_b_2", "a2_a")

        class B(A):
            msg_decorations_unicode = {
                "t_b": "b_u",
                "t_b_2": "b_u",
            }

            msg_decorations_ascii = {
                "t_b": "b_a",
                "t_b_2": "b_a",
            }

        class C(B):
            pass

        a = A()
        b = B()
        c = C()

        assert A.msg_decorations_unicode["t_a"] == "a_u"
        assert A.msg_decorations_ascii["t_a"] == "a_a"
        assert A.msg_decorations_unicode["t_b"] == "a_u"
        assert A.msg_decorations_ascii["t_b"] == "a_a"
        assert A.msg_decorations_unicode["t_b_2"] == "a_u"
        assert A.msg_decorations_ascii["t_b_2"] == "a_a"

        assert a.msg_decorations_unicode["t_a"] == "a2_u"
        assert a.msg_decorations_ascii["t_a"] == "a2_a"
        assert a.msg_decorations_unicode["t_b"] == "a2_u"
        assert a.msg_decorations_ascii["t_b"] == "a2_a"
        assert a.msg_decorations_unicode["t_b_2"] == "a2_u"
        assert a.msg_decorations_ascii["t_b_2"] == "a2_a"

        assert B.msg_decorations_unicode["t_a"] == "a_u"
        assert B.msg_decorations_ascii["t_a"] == "a_a"
        assert B.msg_decorations_unicode["t_b"] == "b_u"
        assert B.msg_decorations_ascii["t_b"] == "b_a"
        assert B.msg_decorations_unicode["t_b_2"] == "b_u"
        assert B.msg_decorations_ascii["t_b_2"] == "b_a"

        assert b.msg_decorations_unicode["t_a"] == "a2_u"
        assert b.msg_decorations_ascii["t_a"] == "a2_a"
        assert b.msg_decorations_unicode["t_b"] == "b_u"
        assert b.msg_decorations_ascii["t_b"] == "b_a"
        assert b.msg_decorations_unicode["t_b_2"] == "a2_u"
        assert b.msg_decorations_ascii["t_b_2"] == "a2_a"

        assert C.msg_decorations_unicode["t_a"] == "a_u"
        assert C.msg_decorations_ascii["t_a"] == "a_a"
        assert C.msg_decorations_unicode["t_b"] == "b_u"
        assert C.msg_decorations_ascii["t_b"] == "b_a"
        assert C.msg_decorations_unicode["t_b_2"] == "b_u"
        assert C.msg_decorations_ascii["t_b_2"] == "b_a"

        assert c.msg_decorations_unicode["t_a"] == "a2_u"
        assert c.msg_decorations_ascii["t_a"] == "a2_a"
        assert c.msg_decorations_unicode["t_b"] == "b_u"
        assert c.msg_decorations_ascii["t_b"] == "b_a"
        assert c.msg_decorations_unicode["t_b_2"] == "a2_u"
        assert c.msg_decorations_ascii["t_b_2"] == "a2_a"


class TestBrokenTheme:
    class RecursiveTheme(yuio.theme.Theme):
        colors = {
            "a": "b",
            "b": "c",
            "c": "a",
        }

    class ThemeEmptyKey(yuio.theme.Theme):
        colors = {
            "": "bold",
        }

    class ThemeEmptyValue(yuio.theme.Theme):
        colors = {
            "bold": "",
        }

    def test_recursion(self):
        theme = self.RecursiveTheme()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", yuio.theme.RecursiveThemeWarning)
            assert theme.get_color("a") == yuio.color.Color.NONE

    def test_warnings(self):
        theme = self.RecursiveTheme()
        with warnings.catch_warnings():
            warnings.simplefilter("error", yuio.theme.RecursiveThemeWarning)
            with pytest.raises(yuio.theme.RecursiveThemeWarning):
                theme.get_color("a")

    def test_check(self):
        theme = self.RecursiveTheme()
        with pytest.raises(yuio.theme.RecursiveThemeWarning):
            theme.check()

    def test_empty_key(self):
        theme = self.ThemeEmptyKey()
        with pytest.warns(
            yuio.theme.ThemeWarning, match=r"colors map contains an empty key"
        ):
            theme.check()

    def test_empty_value(self):
        theme = self.ThemeEmptyValue()
        with pytest.warns(
            yuio.theme.ThemeWarning, match=r"color value for path 'bold' is empty"
        ):
            theme.check()


terminal_theme_dark = yuio.term.TerminalTheme(
    background=yuio.color.ColorValue.from_hex("#00240A"),
    foreground=yuio.color.ColorValue.from_hex("#FFCFCF"),
    black=yuio.color.ColorValue.from_hex("#000000"),
    red=yuio.color.ColorValue.from_hex("#FF0000"),
    green=yuio.color.ColorValue.from_hex("#00FF00"),
    yellow=yuio.color.ColorValue.from_hex("#FFFF00"),
    blue=yuio.color.ColorValue.from_hex("#0000FF"),
    magenta=yuio.color.ColorValue.from_hex("#FF00FF"),
    cyan=yuio.color.ColorValue.from_hex("#00FFFF"),
    white=yuio.color.ColorValue.from_hex("#FFFFFF"),
    lightness=yuio.term.Lightness.DARK,
)

terminal_theme_light = dataclasses.replace(
    terminal_theme_dark,
    background=yuio.color.ColorValue.from_hex("#FFCFCF"),
    foreground=yuio.color.ColorValue.from_hex("#00240A"),
    lightness=yuio.term.Lightness.LIGHT,
)

terminal_theme_unknown = dataclasses.replace(
    terminal_theme_dark,
    lightness=yuio.term.Lightness.UNKNOWN,
)


class TestDefaultTheme:
    def test_no_color_overrides(self):
        term = yuio.term.Term(None, None)  # type: ignore
        theme = yuio.theme.DefaultTheme(term)
        assert (
            theme.get_color("low_priority_color_a") == yuio.color.Color.FORE_NORMAL_DIM
        )
        assert (
            theme.get_color("low_priority_color_b") == yuio.color.Color.FORE_NORMAL_DIM
        )

    def test_no_color_overrides_unknown_lightness(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_theme=terminal_theme_unknown,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert (
            theme.get_color("low_priority_color_a") == yuio.color.Color.FORE_NORMAL_DIM
        )
        assert (
            theme.get_color("low_priority_color_b") == yuio.color.Color.FORE_NORMAL_DIM
        )
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.color.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.color.Color.fore_from_hex("#FF00FF")

    def test_dark_theme(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_theme=terminal_theme_dark,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert theme.get_color(
            "low_priority_color_a"
        ) == yuio.color.Color.fore_from_hex("#655151")
        assert theme.get_color(
            "low_priority_color_b"
        ) == yuio.color.Color.fore_from_hex("#5A4949")
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.color.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.color.Color.fore_from_hex("#FF00FF")

    def test_light_theme(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_theme=terminal_theme_light,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert theme.get_color(
            "low_priority_color_a"
        ) == yuio.color.Color.fore_from_hex("#00B231")
        assert theme.get_color(
            "low_priority_color_b"
        ) == yuio.color.Color.fore_from_hex("#00BF35")
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.color.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.color.Color.fore_from_hex("#FF00FF")

    def test_check(self):
        theme = yuio.theme.DefaultTheme(
            yuio.term.Term(
                None,  # type: ignore
                None,  # type: ignore
            )
        )

        theme.check()


class TestLoad:
    @pytest.fixture(autouse=True)
    def setup_theme_file(self, tmp_path: pathlib.Path):
        self.ok = tmp_path / "ok.json"
        self.ok.write_text(
            textwrap.dedent(
                """
            {
                "progress_bar_width": 25,
                "spinner_update_rate_ms": 100,
                "separate_headings": false,
                "fallback_width": 100,
                "colors": {
                    "red": "#ff5555",
                    "green": "#55ff55",
                    "bold_red": "bold red"
                },
                "msg_decorations_unicode": {
                    "heading/1": "=> "
                }
            }
        """
            )
        )

        self.include = tmp_path / "include.json"
        self.include.write_text(
            textwrap.dedent(
                """
            {
                "include": "ok.json",
                "colors": {
                    "red": "#ff0000"
                }
            }
        """
            )
        )

        self.invalid = tmp_path / "invalid.json"
        self.invalid.write_text(
            textwrap.dedent(
                """
            {
                "progress_bar_width": -1
            }
        """
            )
        )

        self.broken = tmp_path / "broken.json"
        self.broken.write_text(
            textwrap.dedent(
                """
            broken json
        """
            )
        )

    def test_ok(self, term, monkeypatch):
        monkeypatch.setenv("YUIO_THEME_PATH", str(self.ok))
        theme = yuio.theme.load(term)

        assert theme.progress_bar_width == 25
        assert theme.separate_headings is False
        assert theme.fallback_width == 100
        assert theme.spinner_update_rate_ms == 100
        assert theme.get_color("red") == yuio.color.Color.fore_from_rgb(
            0xFF, 0x55, 0x55
        )
        assert theme.get_color("green") == yuio.color.Color.fore_from_rgb(
            0x55, 0xFF, 0x55
        )
        assert theme.get_color("bold_red") == yuio.color.Color.fore_from_rgb(
            0xFF, 0x55, 0x55, bold=True
        )
        assert theme.msg_decorations_unicode["heading/1"] == "=> "

    def test_include(self, term, monkeypatch):
        monkeypatch.setenv("YUIO_THEME_PATH", str(self.include))
        theme = yuio.theme.load(term)

        assert theme.progress_bar_width == 25
        assert theme.separate_headings is False
        assert theme.fallback_width == 100
        assert theme.spinner_update_rate_ms == 100
        assert theme.get_color("red") == yuio.color.Color.fore_from_rgb(
            0xFF, 0x00, 0x00
        )
        assert theme.get_color("green") == yuio.color.Color.fore_from_rgb(
            0x55, 0xFF, 0x55
        )
        assert theme.get_color("bold_red") == yuio.color.Color.fore_from_rgb(
            0xFF, 0x00, 0x00, bold=True
        )
        assert theme.msg_decorations_unicode["heading/1"] == "=> "

    def test_invalid(self, term, monkeypatch):
        monkeypatch.setenv("YUIO_THEME_PATH", str(self.invalid))
        with pytest.warns(yuio.theme.ThemeWarning):
            theme = yuio.theme.load(term)
        assert theme.progress_bar_width == 15

    def test_broken(self, term, monkeypatch):
        monkeypatch.setenv("YUIO_THEME_PATH", str(self.invalid))
        with pytest.warns(yuio.theme.ThemeWarning):
            theme = yuio.theme.load(term)
        assert theme.progress_bar_width == 15



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides base functionality to build CLI interfaces.

Creating and running an app
---------------------------

Yuio's CLI applications have functional interface. Decorate main function
with the :func:`app` decorator, and use :meth:`App.run` method to start it:

.. code-block:: python

    # Let's define an app with one flag and one positional argument.
    @app
    def main(
        #: help message for `arg`
        arg: str = positional(),
        #: help message for `--flag`
        flag: int = 0
    ):
        \"""this command does a thing\"""
        yuio.io.info("flag=%r, arg=%r", flag, arg)

    if __name__ == "__main__":
        # We can now use `main.run` to parse arguments and invoke `main`.
        # Notice that `run` does not return anything. Instead, it terminates
        # python process with an appropriate exit code.
        main.run("--flag 10 foobar!".split())

Function's arguments will become program's flags and positionals, and function's
docstring will become app's :attr:`~App.description`.

Help messages for the flags are parsed from line comments
right above the field definition (comments must start with ``#:``).
They are all formatted using Markdown (see :mod:`yuio.md`).

Parsers for CLI argument values are derived from type hints.
Use the `parser` parameter of the :func:`field` function to override them.

Arguments with bool parsers and parsers that support
:meth:`parsing collections <yuio.parse.Parser.supports_parse_many>`
are handled to provide better CLI experience:

.. invisible-code-block: python

    import pathlib

.. code-block:: python

    @app
    def main(
        # Will create flags `--verbose` and `--no-verbose`.
        # Since default is `False`, `--no-verbose` will be hidden from help
        # to reduce clutter.
        verbose: bool = False,
        # Will create a flag with `nargs=*`: `--inputs path1 path2 ...`
        inputs: list[pathlib.Path] = [],
    ): ...

.. autofunction:: app

.. autoclass:: App

    .. automethod:: run

    .. method:: wrapped(...)

        The original callable what was wrapped by :func:`app`.


Configuring CLI arguments
-------------------------

Names and types of arguments are determined by names and types of the app function's
arguments. You can use the :func:`field` function to override them:

.. autofunction:: field

.. autofunction:: inline

.. autofunction:: positional


Creating argument groups
------------------------

You can use :class:`~yuio.config.Config` as a type of an app function's parameter.
This will make all of config fields into flags as well. By default, Yuio will use
parameter name as a prefix for all fields in the config; you can override it
with :func:`field` or :func:`inline`:

.. code-block:: python

    class KillCmdConfig(yuio.config.Config):
        # Will be loaded from `--signal`.
        signal: int

        # Will be loaded from `-p` or `--pid`.
        pid: int = field(flags=["-p", "--pid"])


    @app
    def main(
        # `kill_cmd.signal` will be loaded from `--kill-cmd-signal`.
        kill_cmd: KillCmdConfig,
        # `copy_cmd_2.signal` will be loaded from `--kill-signal`.
        kill_cmd_2: KillCmdConfig = field(flags="--kill"),
        # `kill_cmd_3.signal` will be loaded from `--signal`.
        kill_cmd_3: KillCmdConfig = field(flags=""),
    ): ...

.. note::

    Positional arguments are not allowed in configs,
    only in apps.


App settings
------------

You can override default usage and help messages as well as control some of the app's
help formatting using its arguments:

.. class:: App
    :noindex:

    .. autoattribute:: prog

    .. autoattribute:: usage

    .. autoattribute:: description

    .. autoattribute:: help

    .. autoattribute:: epilog

    .. autoattribute:: allow_abbrev

    .. autoattribute:: subcommand_required

    .. autoattribute:: setup_logging

    .. autoattribute:: theme

    .. autoattribute:: version

    .. autoattribute:: bug_report

    .. autoattribute:: is_dev_mode


Creating sub-commands
---------------------

You can create multiple sub-commands for the main function
using the :meth:`App.subcommand` method:

.. code-block:: python

    @app
    def main(): ...


    @main.subcommand
    def do_stuff(): ...

There is no limit to how deep you can nest subcommands, but for usability reasons
we suggest not exceeding level of sub-sub-commands (:flag:`git stash push`, anyone?)

When user invokes a subcommand, the ``main()`` function is called first,
then subcommand. In the above example, invoking our app with subcommand ``push``
will cause ``main()`` to be called first, then ``push()``.

This behavior is useful when you have some global configuration flags
attached to the ``main()`` command. See the `example app`_ for details.

.. _example app: https://github.com/taminomara/yuio/blob/main/examples/app

.. class:: App
    :noindex:

    .. automethod:: subcommand


.. _sub-commands-more:

Controlling how sub-commands are invoked
----------------------------------------

By default, if a command has sub-commands, the user is required to provide
a sub-command. This behavior can be disabled by setting :attr:`App.subcommand_required`
to :data:`False`.

When this happens, we need to understand whether a subcommand was invoked or not.
To determine this, you can accept a special parameter called `_command_info`
of type :class:`CommandInfo`. It will contain info about the current function,
including its name and subcommand:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        if _command_info.subcommand is not None:
            # A subcommand was invoked.
            ...

You can call the subcommand on your own by using ``_command_info.subcommand``
as a callable:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        if _command_info.subcommand is not None and ...:
            _command_info.subcommand()  # manually invoking a subcommand

If you wish to disable calling the subcommand, you can return :data:`False`
from the main function:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        ...
        # Subcommand will not be invoked.
        return False

.. autoclass:: CommandInfo
   :members:


.. _flags-with-multiple-values:

Handling options with multiple values
-------------------------------------

When you create an option with a container type, Yuio enables passing its values
by specifying multiple arguments. For example:

.. code-block:: python

    @yuio.app.app
    def main(list: list[int]):
        print(list)

Here, you can pass values to :flag:`--list` as separate arguments:

.. code-block:: console

    $ app --list 1 2 3
    [1, 2, 3]

If you specify value for :flag:`--list` inline, it will be handled as
a delimiter-separated list:

.. code-block:: console

    $ app --list='1 2 3'
    [1, 2, 3]

This allows resolving ambiguities between flags and positional arguments:

.. code-block:: console

    $ app --list='1 2 3' subcommand

Technically, :flag:`--list 1 2 3` causes Yuio to invoke
``list_parser.parse_many(["1", "2", "3"])``, while :flag:`--list='1 2 3'` causes Yuio
to invoke ``list_parser.parse("1 2 3")``.


.. _flags-with-optional-values:

Handling flags with optional values
-----------------------------------

When designing a CLI, one important question is how to handle flags with optional
values, if at all. There are several things to consider:

1.  Does a flag have clear and predictable behavior when its value is not specified?

    For boolean flags the default behavior is obvious: :flag:`--use-gpu` will enable
    GPU, i.e. it is equivalent to :flag:`--use-gpu=true`.

    For flags that accept non-boolean values, though, things get messier. What will
    a flag like :flag:`--n-threads` do? Will it calculate number of threads based on
    available CPU cores? Will it use some default value?

    In these cases, it is usually better to require a sentinel value:
    :flag:`--n-threads=auto`.

2.  Where should flag's value go, it it's provided?

    We can only allow passing value inline, i.e. :flag:`--use-gpu=true`. Or we can
    greedily take the following argument as flag's value, i.e. :flag:`--use-gpu true`.

    The later approach has a significant downside: we don't know
    whether the next argument was intended for the flag or for a free-standing option.

    For example:

    .. code-block:: console

        $ my-renderer --color true  # is `true` meant for `--color`,
        $                           # or is it a subcommand for `my-renderer`?

Here's how Yuio handles this dilemma:

1.  High level API does not allow creating flags with optional values.

    To create one, you have to make a custom implementation of :class:`yuio.cli.Option`
    and set its :attr:`~yuio.cli.Option.allow_no_args` to :data:`True`. This will
    correspond to the greedy approach.

    .. note::

        Positionals with defaults are treated as optional because they don't
        create ambiguities.

2.  Boolean flags allow specifying value inline, but not as a separate argument.

3.  Yuio does not allow passing inline values to short boolean flags
    without adding an equals sign. For example, :flag:`-ftrue` will not work,
    while :flag:`-f=true` will.

    This is done to enable grouping short flags: :flag:`ls -laH` should be parsed
    as :flag:`ls -l -a -H`, not as :flag:`ls -l=aH`.

4.  On lower levels of API, Yuio allows precise control over these behavior
    by setting :attr:`Option.nargs <yuio.cli.Option.nargs>`,
    :attr:`Option.allow_no_args <yuio.cli.Option.allow_no_args>`,
    :attr:`Option.allow_inline_arg <yuio.cli.Option.allow_inline_arg>`,
    and :attr:`Option.allow_implicit_inline_arg <yuio.cli.Option.allow_implicit_inline_arg>`.


.. _custom-cli-options:

Creating custom CLI options
---------------------------

You can override default behavior and presentation of a CLI option by passing
custom `option_ctor` to :func:`field`. Furthermore, you can create your own
implementation of :class:`yuio.cli.Option` to further fine-tune how an option
is parsed, presented in CLI help, etc.

.. autofunction:: bool_option

.. autofunction:: parse_one_option

.. autofunction:: parse_many_option

.. autofunction:: store_const_option

.. autofunction:: count_option

.. autofunction:: store_true_option

.. autofunction:: store_false_option

.. type:: OptionCtor
    :canonical: typing.Callable[[OptionSettings], yuio.cli.Option[T]]

    CLI option constructor. Takes a single positional argument
    of type :class:`OptionSettings`, and returns an instance
    of :class:`yuio.cli.Option`.

.. autoclass:: OptionSettings
    :members:


Re-imports
----------

.. type:: HelpGroup
    :no-index:

    Alias of :obj:`yuio.cli.HelpGroup`.

.. type:: MutuallyExclusiveGroup
    :no-index:

    Alias of :obj:`yuio.cli.MutuallyExclusiveGroup`.

.. data:: MISC_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.MISC_GROUP`.

.. data:: OPTS_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.OPTS_GROUP`.

.. data:: SUBCOMMANDS_GROUP
    :no-index:

    Alias of :obj:`yuio.cli.SUBCOMMANDS_GROUP`.

"""

from __future__ import annotations

import dataclasses
import functools
import inspect
import json
import logging
import pathlib
import sys
import types
from dataclasses import dataclass

import yuio
import yuio.cli
import yuio.complete
import yuio.config
import yuio.dbg
import yuio.io
import yuio.parse
import yuio.string
import yuio.term
import yuio.theme
import yuio.util
from yuio.cli import (
    MISC_GROUP,
    OPTS_GROUP,
    SUBCOMMANDS_GROUP,
    HelpGroup,
    MutuallyExclusiveGroup,
)
from yuio.config import (
    OptionCtor,
    OptionSettings,
    bool_option,
    count_option,
    field,
    inline,
    parse_many_option,
    parse_one_option,
    positional,
    store_const_option,
    store_false_option,
    store_true_option,
)
from yuio.util import _find_docs
from yuio.util import to_dash_case as _to_dash_case

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

__all__ = [
    "MISC_GROUP",
    "OPTS_GROUP",
    "SUBCOMMANDS_GROUP",
    "App",
    "AppError",
    "CommandInfo",
    "HelpGroup",
    "MutuallyExclusiveGroup",
    "OptionCtor",
    "OptionSettings",
    "app",
    "bool_option",
    "count_option",
    "field",
    "inline",
    "parse_many_option",
    "parse_one_option",
    "positional",
    "store_const_option",
    "store_false_option",
    "store_true_option",
]

C = _t.TypeVar("C", bound=_t.Callable[..., None | bool])
C2 = _t.TypeVar("C2", bound=_t.Callable[..., None | bool])


class AppError(yuio.PrettyException, Exception):
    """
    An error that you can throw from an app to finish its execution without printing
    a traceback.

    """


@_t.overload
def app(
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    version: str | None = None,
    bug_report: yuio.dbg.ReportSettings | bool = False,
    is_dev_mode: bool | None = None,
) -> _t.Callable[[C], App[C]]: ...
@_t.overload
def app(
    command: C,
    /,
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    version: str | None = None,
    bug_report: yuio.dbg.ReportSettings | bool = False,
    is_dev_mode: bool | None = None,
) -> App[C]: ...
def app(
    command: _t.Callable[..., None | bool] | None = None,
    /,
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    allow_abbrev: bool = False,
    subcommand_required: bool = True,
    setup_logging: bool = True,
    theme: (
        yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
    ) = None,
    version: str | None = None,
    bug_report: yuio.dbg.ReportSettings | bool = False,
    is_dev_mode: bool | None = None,
) -> _t.Any:
    """
    Create an application.

    This is a decorator that's supposed to be used on the main method
    of the application. This decorator returns an :class:`App` object.

    :param command:
        the main function of the application.
    :param prog:
        overrides program's name, see :attr:`App.prog`.
    :param usage:
        overrides program's usage description, see :attr:`App.usage`.
    :param description:
        overrides program's description, see :attr:`App.description`.
    :param epilog:
        overrides program's epilog, see :attr:`App.epilog`.
    :param allow_abbrev:
        whether to allow abbreviating unambiguous flags, see :attr:`App.allow_abbrev`.
    :param subcommand_required:
        whether this app requires a subcommand,
        see :attr:`App.subcommand_required`.
    :param setup_logging:
        whether to perform basic logging setup on startup,
        see :attr:`App.setup_logging`.
    :param theme:
        overrides theme that will be used when setting up :mod:`yuio.io`,
        see :attr:`App.theme`.
    :param version:
        program's version, will be displayed using the :flag:`--version` flag.
    :param bug_report:
        settings for automated bug report generation. If present,
        adds the :flag:`--bug-report` flag.
    :param is_dev_mode:
        enables additional logging, see :attr:`App.is_dev_mode`.
    :returns:
        an :class:`App` object that wraps the original function.

    """

    def registrar(command: C, /) -> App[C]:
        return App(
            command,
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            allow_abbrev=allow_abbrev,
            subcommand_required=subcommand_required,
            setup_logging=setup_logging,
            theme=theme,
            version=version,
            bug_report=bug_report,
            is_dev_mode=is_dev_mode,
        )

    if command is None:
        return registrar
    else:
        return registrar(command)


@_t.final
@dataclass(frozen=True, eq=False, match_args=False, slots=True)
class CommandInfo:
    """
    Data about the invoked command.

    """

    name: str
    """
    Name of the current command.

    If it was invoked by alias,
    this will contains the primary command name.

    For the main function, the name will be set to ``"__main__"``.

    """

    # Internal, do not use.
    _config: _t.Any = dataclasses.field(repr=False)
    _executed: bool = dataclasses.field(default=False, repr=False)
    _subcommand: CommandInfo | None | yuio.Missing = dataclasses.field(
        default=yuio.MISSING, repr=False
    )

    @property
    def subcommand(self) -> CommandInfo | None:
        """
        Subcommand of this command, if one was given.

        """

        if self._subcommand is yuio.MISSING:
            if self._config._subcommand is None:
                subcommand = None
            else:
                subcommand = CommandInfo(
                    self._config._subcommand, self._config._subcommand_ns.config
                )
            object.__setattr__(self, "_subcommand", subcommand)
        return self._subcommand  # pyright: ignore[reportReturnType]

    def __call__(self) -> _t.Literal[False]:
        """
        Execute this command.

        """

        if self._executed:
            return False
        object.__setattr__(self, "_executed", True)

        should_invoke_subcommand = self._config._run(self)
        if should_invoke_subcommand is None:
            should_invoke_subcommand = True

        if should_invoke_subcommand and self.subcommand is not None:
            self.subcommand()

        return False


class App(_t.Generic[C]):
    """
    A class that encapsulates app settings and logic for running it.

    It is better to create instances of this class using the :func:`app` decorator,
    as it provides means to decorate the main function and specify all of the app's
    parameters.

    """

    @dataclass(frozen=True, eq=False, match_args=False, slots=True)
    class _SubApp:
        app: App[_t.Any]
        name: str
        aliases: list[str] | None = None
        is_primary: bool = False

    def __init__(
        self,
        command: C,
        /,
        *,
        prog: str | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
        allow_abbrev: bool = False,
        subcommand_required: bool = True,
        setup_logging: bool = True,
        theme: (
            yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
        ) = None,
        version: str | None = None,
        bug_report: yuio.dbg.ReportSettings | bool = False,
        is_dev_mode: bool | None = None,
    ):
        self.prog: str | None = prog
        """
        Program or subcommand's primary name.

        For main app, this controls its display name and generation of shell completion
        scripts.

        For subcommands, this is always equal to subcommand's main name.

        By default, inferred from :data:`sys.argv` and subcommand name.

        """

        self.usage: str | None = usage
        """
        Program or subcommand synapsis.

        This string will be processed using the to ``bash`` syntax,
        and then it will be ``%``-formatted with a single keyword argument ``prog``.
        If command supports multiple signatures, each of them should be listed
        on a separate string. For example::

            @app
            def main(): ...

            main.usage = \"""
            %(prog)s [-q] [-f] [-m] [<branch>]
            %(prog)s [-q] [-f] [-m] --detach [<branch>]
            %(prog)s [-q] [-f] [-m] [--detach] <commit>
            ...
            \"""

        By default, usage is generated from CLI flags.

        See `usage <https://docs.python.org/3/library/argparse.html#usage>`_
        in :mod:`argparse`.

        """

        if not description and command.__doc__:
            description = yuio.util._process_docstring(
                command.__doc__, only_first_paragraph=False
            )

        self.description: str | None = description
        """
        Text that is shown before CLI flags help, usually contains
        short description of the program or subcommand.

        The text should be formatted using markdown. For example:

        .. code-block:: python

           @app
           def main(): ...

           main.description = \"""
           This command does a thing.

           # Different ways to do a thing:

           This command can apply multiple algorithms to achieve
           a necessary state in which a thing can be done. This includes:

           - randomly turning the screen on and off;

           - banging a head on a table;

           - fiddling with your PCs power cord.

           By default, the best algorithm is determined automatically.
           However, you can hint a preferred algorithm via the `--hint-algo` flag.

           \"""

        By default, inferred from command's docstring.

        """

        if help is None and description:
            help = description
            if (index := help.find("\n\n")) != -1:
                help = help[:index]

        self.help: str | yuio.Disabled | None = help
        """
        Short help message that is shown when listing subcommands.

        By default, uses first paragraph of description.

        """

        self.epilog: str | None = epilog
        """
        Text that is shown after the main portion of the help message.

        Text format is identical to the one for :attr:`~App.description`.

        """

        self.allow_abbrev: bool = allow_abbrev
        """
        Allow abbreviating CLI flags if that doesn't create ambiguity.

        Disabled by default.

        """

        self.subcommand_required: bool = subcommand_required
        """
        Require the user to provide a subcommand for this command.

        If this command doesn't have any subcommands, this option is ignored.

        Enabled by default.

        """

        self.setup_logging: bool = setup_logging
        """
        If :data:`True`, the app will call :func:`logging.basicConfig` during
        its initialization. Disable this if you want to customize
        logging initialization.

        Disabling this option also removes the :flag:`--verbose` flag form the CLI.

        """

        self.theme: (
            yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
        ) = theme
        """
        A custom theme that will be passed to :func:`yuio.io.setup`
        on application startup.

        """

        self.version: str | None = version
        """
        If not :data:`None`, add :flag:`--version` flag to the CLI.

        """

        self.bug_report: yuio.dbg.ReportSettings | bool = bug_report
        """
        If not :data:`False`, add :flag:`--bug-report` flag to the CLI.

        This flag automatically collects data about environment and prints it
        in a format suitable for adding to a bug report.

        """

        self.is_dev_mode: bool | None = is_dev_mode
        """
        If :data:`True`, this will enable :func:`logging.captureWarnings`
        and configure internal Yuio logging to show warnings.

        By default, dev mode is detected by checking if :attr:`~App.version`
        contains substring ``"dev"``.

        .. note::

            You can always enable full debug logging by setting environment
            variable ``YUIO_DEBUG``.

            If enabled, full log will be saved to ``YUIO_DEBUG_FILE``.

        """

        self.__sub_apps: dict[str, App._SubApp] = {}

        if callable(command):
            self.__config_type = _command_from_callable(command)
        else:
            raise TypeError(f"expected a function, got {command}")

        functools.update_wrapper(
            self,  # type: ignore
            command,
            assigned=("__module__", "__name__", "__qualname__", "__doc__"),
            updated=(),
        )

        self._command = command

        @functools.wraps(command)
        def wrapped_command(*args, **kwargs):
            if args:
                names = self.__config_type.__annotations__
                if len(args) > len(names):
                    s = "" if len(names) == 1 else "s"
                    raise TypeError(
                        f"expected at most {len(names)} positional argument{s}, got {len(args)}"
                    )
                for arg, name in zip(args, names):
                    if name in kwargs:
                        raise TypeError(f"argument {name} was given twice")
                    kwargs[name] = arg
            return CommandInfo("__raw__", self.__config_type(**kwargs), False)()

        self.wrapped: C = wrapped_command  # type: ignore
        """
        The original callable what was wrapped by :func:`app`.

        """

    @_t.overload
    def subcommand(
        self,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
    ) -> _t.Callable[[C2], App[C2]]: ...

    @_t.overload
    def subcommand(
        self,
        cb: C2,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
    ) -> App[C2]: ...

    def subcommand(
        self,
        cb: _t.Callable[..., None | bool] | None = None,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
        subcommand_required: bool = True,
    ) -> _t.Any:
        """
        Register a subcommand for the given app.

        This method can be used as a decorator, similar to the :func:`app` function.

        :param name:
            allows overriding subcommand's name.
        :param aliases:
            allows adding alias names for subcommand.
        :param usage:
            overrides subcommand's usage description, see :attr:`App.usage`.
        :param help:
            overrides subcommand's short help, see :attr:`App.help`.
            pass :data:`~yuio.DISABLED` to hide this subcommand in CLI help message.
        :param description:
            overrides subcommand's description, see :attr:`App.description`.
        :param epilog:
            overrides subcommand's epilog, see :attr:`App.epilog`.
        :param subcommand_required:
            whether this subcommand requires another subcommand,
            see :attr:`App.subcommand_required`.
        :returns:
            a new :class:`App` object for a subcommand.

        """

        def registrar(cb: C2, /) -> App[C2]:
            main_name = name or _to_dash_case(cb.__name__)
            app = App(
                cb,
                prog=main_name,
                usage=usage,
                help=help,
                description=description,
                epilog=epilog,
                subcommand_required=subcommand_required,
            )

            self.__sub_apps[main_name] = App._SubApp(
                app, main_name, aliases, is_primary=True
            )
            if aliases:
                alias_app = App._SubApp(app, main_name)
                self.__sub_apps.update({alias: alias_app for alias in aliases})

            return app

        if cb is None:
            return registrar
        else:
            return registrar(cb)

    def run(self, args: list[str] | None = None) -> _t.NoReturn:
        """
        Parse arguments, set up :mod:`yuio.io` and :mod:`logging`,
        and run the application.

        :param args:
            command line arguments. If none are given,
            use arguments from :data:`sys.argv`.
        :returns:
            this method does not return, it exits the program instead.

        """

        if args is None:
            args = sys.argv[1:]

        if "--yuio-custom-completer--" in args:
            index = args.index("--yuio-custom-completer--")
            _run_custom_completer(
                self.__make_cli_command(root=True), args[index + 1], args[index + 2]
            )
            sys.exit(0)

        if "--yuio-bug-report--" in args:
            from yuio.dbg import print_report

            print_report(settings=self.bug_report, app=self)
            sys.exit(0)

        yuio.io.setup(theme=self.theme, wrap_stdio=True)

        try:
            if self.is_dev_mode is None:
                self.is_dev_mode = (
                    self.version is not None and "dev" in self.version.casefold()
                )
            if self.is_dev_mode:
                yuio.enable_internal_logging(add_handler=True)

            cli_command = self.__make_cli_command(root=True)
            namespace = yuio.cli.CliParser(cli_command).parse(args)

            if self.setup_logging:
                logging_level = {
                    0: logging.WARNING,
                    1: logging.INFO,
                    2: logging.DEBUG,
                }.get(namespace["_verbosity"], logging.DEBUG)
                logging.basicConfig(handlers=[yuio.io.Handler()], level=logging_level)

            command = CommandInfo("__main__", _config=namespace.config)
            command()
            sys.exit(0)
        except yuio.cli.ArgumentError as e:
            yuio.io.raw(e, add_newline=True)
            sys.exit(1)
        except (AppError, yuio.cli.ArgumentError, yuio.parse.ParsingError) as e:
            yuio.io.failure(e)
            sys.exit(1)
        except KeyboardInterrupt:
            yuio.io.failure("Received Keyboard Interrupt, stopping now")
            sys.exit(130)
        except Exception as e:
            yuio.io.failure_with_tb("Error: %s", e)
            sys.exit(3)
        finally:
            yuio.io.restore_streams()

    def __make_cli_command(self, root: bool = False):
        options = self.__config_type._build_options()

        if root:
            options.append(yuio.cli.HelpOption())
            if self.version:
                options.append(yuio.cli.VersionOption(version=self.version))
            if self.setup_logging:
                options.append(
                    yuio.cli.CountOption(
                        flags=["-v", "--verbose"],
                        usage=yuio.GROUP,
                        help="Increase output verbosity.",
                        help_group=yuio.cli.MISC_GROUP,
                        show_if_inherited=False,
                        dest="_verbosity",
                    )
                )
            if self.bug_report:
                options.append(yuio.cli.BugReportOption(app=self))
            options.append(yuio.cli.CompletionOption())
            options.append(_ColorOption())

        subcommands = {}
        for sub_app in self.__sub_apps.values():
            if not sub_app.is_primary:
                continue
            aubcommand = sub_app.app.__make_cli_command()
            subcommands[sub_app.name] = aubcommand
            for alias in sub_app.aliases or []:
                subcommands[alias] = aubcommand

        return yuio.cli.Command(
            name=self.prog or pathlib.Path(sys.argv[0]).stem,
            desc=self.description or "",
            help=self.help if self.help is not None else "",
            epilog=self.epilog or "",
            usage=self.usage,
            options=options,
            subcommands=subcommands,
            subcommand_required=self.subcommand_required,
            ns_ctor=lambda: yuio.cli.ConfigNamespace(self.__config_type()),
            dest="_subcommand",
            ns_dest="_subcommand_ns",
        )


def _command_from_callable(
    cb: _t.Callable[..., None | bool],
) -> type[yuio.config.Config]:
    sig = inspect.signature(cb)

    dct = {}
    annotations = {}

    accepts_command_info = False

    try:
        docs = _find_docs(cb)
    except Exception:
        yuio._logger.warning(
            "unable to get documentation for %s.%s",
            cb.__module__,
            cb.__qualname__,
        )
        docs = {}

    for name, param in sig.parameters.items():
        if param.kind not in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
            raise TypeError("positional-only and variadic arguments are not supported")

        if name.startswith("_"):
            if name == "_command_info":
                accepts_command_info = True
                continue
            else:
                raise TypeError(f"unknown special param {name}")

        if param.default is not param.empty:
            field = param.default
        else:
            field = yuio.MISSING
        if not isinstance(field, yuio.config._FieldSettings):
            field = _t.cast(
                yuio.config._FieldSettings, yuio.config.field(default=field)
            )
        if name in docs:
            field = dataclasses.replace(field, help=docs[name])

        if param.annotation is param.empty:
            raise TypeError(f"param {name} requires type annotation")

        dct[name] = field
        annotations[name] = param.annotation

    dct["_run"] = _command_from_callable_run_impl(
        cb, list(annotations.keys()), accepts_command_info
    )
    dct["_color"] = None
    dct["_verbosity"] = 0
    dct["_subcommand"] = None
    dct["_subcommand_ns"] = None
    dct["__annotations__"] = annotations
    dct["__module__"] = getattr(cb, "__module__", None)
    dct["__doc__"] = getattr(cb, "__doc__", None)

    return types.new_class(
        cb.__name__,
        (yuio.config.Config,),
        {"_allow_positionals": True},
        exec_body=lambda ns: ns.update(dct),
    )


def _command_from_callable_run_impl(
    cb: _t.Callable[..., None | bool], params: list[str], accepts_command_info
):
    def run(self, command_info):
        kw = {name: getattr(self, name) for name in params}
        if accepts_command_info:
            kw["_command_info"] = command_info
        return cb(**kw)

    return run


def _run_custom_completer(command: yuio.cli.Command[_t.Any], raw_data: str, word: str):
    data = json.loads(raw_data)
    path: str = data["path"]
    flags: set[str] = set(data["flags"])
    index: int = data["index"]

    root = command
    for name in path.split("/"):
        if not name:
            continue
        if name not in command.subcommands:
            return
        root = command.subcommands[name]

    positional_index = 0
    for option in root.options:
        option_flags = option.flags
        if option_flags is yuio.POSITIONAL:
            option_flags = [str(positional_index)]
            positional_index += 1
        if flags.intersection(option_flags):
            completer, is_many = option.get_completer()
            break
    else:
        completer, is_many = None, False

    if completer:
        yuio.complete._run_completer_at_index(completer, is_many, index, word)


@dataclass(eq=False, kw_only=True)
class _ColorOption(yuio.cli.Option[_t.Never]):
    # `yuio.term` will scan `sys.argv` on its own, this option just checks format
    # and adds help entry.

    _ALLOWED_VALUES = (
        "y",
        "yes",
        "true",
        "1",
        "n",
        "no",
        "false",
        "0",
        "ansi",
        "ansi-256",
        "ansi-true",
    )

    _PUBLIC_VALUES = (
        ("true", "3-bit colors or higher"),
        ("false", "disable colors"),
        ("ansi", "force 3-bit colors"),
        ("ansi-256", "force 8-bit colors"),
        ("ansi-true", "force 24-bit colors"),
    )

    def __init__(self):
        super().__init__(
            flags=["--color", "--no-color"],
            allow_inline_arg=True,
            allow_implicit_inline_arg=True,
            nargs=0,
            allow_no_args=True,
            required=False,
            metavar=(),
            mutex_group=None,
            usage=yuio.GROUP,
            help="Enable or disable ANSI colors.",
            help_group=yuio.cli.MISC_GROUP,
            show_if_inherited=False,
            allow_abbrev=False,
        )

    def process(
        self,
        cli_parser: yuio.cli.CliParser[yuio.cli.Namespace],
        flag: yuio.cli.Flag | None,
        arguments: yuio.cli.Argument | list[yuio.cli.Argument],
        ns: yuio.cli.Namespace,
    ):
        if isinstance(arguments, yuio.cli.Argument):
            if flag and flag.value == "--no-color":
                raise yuio.cli.ArgumentError(
                    "This flag can't have arguments", flag=flag, arguments=arguments
                )
            if arguments.value.casefold() not in self._ALLOWED_VALUES:
                raise yuio.cli.ArgumentError(
                    "Can't parse `%r` as color, should be %s",
                    arguments.value,
                    yuio.string.Or(value for value, _ in self._PUBLIC_VALUES),
                    flag=flag,
                    arguments=arguments,
                )

    @functools.cached_property
    def primary_short_flag(self):
        return None

    @functools.cached_property
    def primary_long_flags(self):
        return ["--color", "--no-color"]

    def format_alias_flags(
        self,
        ctx: yuio.string.ReprContext,
        /,
        *,
        all: bool = False,
    ) -> list[yuio.string.ColorizedString] | None:
        if self.flags is yuio.POSITIONAL:
            return None

        primary_flags = set(self.primary_long_flags or [])
        if self.primary_short_flag:
            primary_flags.add(self.primary_short_flag)

        aliases: list[yuio.string.ColorizedString] = []
        flag_color = ctx.get_color("hl/flag:sh-usage")
        punct_color = ctx.get_color("hl/punct:sh-usage")
        metavar_color = ctx.get_color("hl/metavar:sh-usage")
        res = yuio.string.ColorizedString()
        res.start_no_wrap()
        res.append_color(flag_color)
        res.append_str("--color")
        res.end_no_wrap()
        res.append_color(punct_color)
        res.append_str("={")
        sep = False
        for value, _ in self._PUBLIC_VALUES:
            if sep:
                res.append_color(punct_color)
                res.append_str("|")
            res.append_color(metavar_color)
            res.append_str(value)
            sep = True
        res.append_color(punct_color)
        res.append_str("}")
        aliases.append(res)
        return aliases

    def get_completer(self) -> tuple[yuio.complete.Completer | None, bool]:
        return yuio.complete.Choice(
            [
                yuio.complete.Option(value, comment)
                for value, comment in self._PUBLIC_VALUES
            ]
        ), False



import yuio.secret


def test_basics():
    s = yuio.secret.SecretValue("asd")
    assert str(s) == "***"
    assert repr(s) == "SecretValue(***)"

    s = yuio.secret.SecretValue("")
    assert str(s) == "***"
    assert repr(s) == "SecretValue(***)"



import yuio.color

import pytest


@pytest.mark.parametrize(
    ("color", "expect"),
    [
        (yuio.color.Color.NONE | yuio.color.Color.NONE, yuio.color.Color.NONE),
        (yuio.color.Color.NONE | yuio.color.Color.FORE_RED, yuio.color.Color.FORE_RED),
        (yuio.color.Color.NONE | yuio.color.Color.BACK_RED, yuio.color.Color.BACK_RED),
        (
            yuio.color.Color.FORE_BLUE | yuio.color.Color.FORE_RED,
            yuio.color.Color.FORE_RED,
        ),
        (
            yuio.color.Color.BACK_BLUE | yuio.color.Color.BACK_RED,
            yuio.color.Color.BACK_RED,
        ),
        (
            yuio.color.Color.FORE_RED | yuio.color.Color.BACK_BLUE,
            yuio.color.Color(
                fore=yuio.color.ColorValue(1), back=yuio.color.ColorValue(4)
            ),
        ),
        (
            yuio.color.Color.FORE_RED | yuio.color.Color.STYLE_BOLD,
            yuio.color.Color(fore=yuio.color.ColorValue(1), bold=True),
        ),
        (
            yuio.color.Color.STYLE_BOLD | yuio.color.Color.FORE_RED,
            yuio.color.Color(fore=yuio.color.ColorValue(1), bold=True),
        ),
        (
            yuio.color.Color.FORE_RED | yuio.color.Color.STYLE_DIM,
            yuio.color.Color(fore=yuio.color.ColorValue(1), dim=True),
        ),
        (
            yuio.color.Color.STYLE_DIM | yuio.color.Color.FORE_RED,
            yuio.color.Color(fore=yuio.color.ColorValue(1), dim=True),
        ),
        (
            yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.STYLE_DIM,
            yuio.color.Color(fore=yuio.color.ColorValue(1), bold=True, dim=True),
        ),
        (
            yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.STYLE_NORMAL,
            yuio.color.Color(
                fore=yuio.color.ColorValue(1),
                bold=False,
                dim=False,
                italic=False,
                underline=False,
                inverse=False,
                blink=False,
            ),
        ),
        (
            yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_NORMAL,
            yuio.color.Color(
                fore=yuio.color.ColorValue(1),
                bold=False,
                dim=False,
                italic=False,
                underline=False,
                inverse=False,
                blink=False,
            ),
        ),
        (
            yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_NORMAL
            | yuio.color.Color.STYLE_DIM,
            yuio.color.Color(
                fore=yuio.color.ColorValue(1),
                bold=False,
                dim=True,
                italic=False,
                underline=False,
                inverse=False,
                blink=False,
            ),
        ),
        (
            yuio.color.Color.FORE_RED
            | yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_NORMAL
            | yuio.color.Color.STYLE_BOLD,
            yuio.color.Color(
                fore=yuio.color.ColorValue(1),
                bold=True,
                dim=False,
                italic=False,
                underline=False,
                inverse=False,
                blink=False,
            ),
        ),
    ],
)
def test_combine(color, expect):
    assert color == expect


@pytest.mark.parametrize(
    ("colors", "coeffs", "expect"),
    [
        (
            [yuio.color.Color.FORE_RED, yuio.color.Color.FORE_GREEN],
            [i / 4 for i in range(5)],
            [yuio.color.Color.FORE_RED] * 5,
        ),
        (
            [yuio.color.Color.FORE_RED, yuio.color.Color.fore_from_hex("#AA0000")],
            [i / 4 for i in range(5)],
            [yuio.color.Color.FORE_RED] * 5,
        ),
        (
            [yuio.color.Color.fore_from_hex("#AA0000"), yuio.color.Color.FORE_RED],
            [i / 4 for i in range(5)],
            [yuio.color.Color.fore_from_hex("#AA0000")] * 5,
        ),
        (
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
            ],
            [i / 4 for i in range(5)],
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#AA0000"),
            ],
        ),
        (
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#00AA00"),
            ],
            [i / 4 for i in range(5)],
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#7F2A00"),
                yuio.color.Color.fore_from_hex("#555500"),
                yuio.color.Color.fore_from_hex("#2A7F00"),
                yuio.color.Color.fore_from_hex("#00AA00"),
            ],
        ),
        (
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#00AA00"),
                yuio.color.Color.fore_from_hex("#0000AA"),
            ],
            [i / 8 for i in range(9)],
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#7F2A00"),
                yuio.color.Color.fore_from_hex("#555500"),
                yuio.color.Color.fore_from_hex("#2A7F00"),
                yuio.color.Color.fore_from_hex("#00AA00"),
                yuio.color.Color.fore_from_hex("#007F2A"),
                yuio.color.Color.fore_from_hex("#005555"),
                yuio.color.Color.fore_from_hex("#002A7F"),
                yuio.color.Color.fore_from_hex("#0000AA"),
            ],
        ),
        (
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#00AA00"),
            ],
            [0, 0.5, 1],
            [
                yuio.color.Color.fore_from_hex("#AA0000"),
                yuio.color.Color.fore_from_hex("#555500"),
                yuio.color.Color.fore_from_hex("#00AA00"),
            ],
        ),
        (
            [
                yuio.color.Color.back_from_hex("#0000AA"),
                yuio.color.Color.back_from_hex("#00AA00"),
            ],
            [0, 0.5, 1],
            [
                yuio.color.Color.back_from_hex("#0000AA"),
                yuio.color.Color.back_from_hex("#005555"),
                yuio.color.Color.back_from_hex("#00AA00"),
            ],
        ),
        (
            [
                yuio.color.Color.fore_from_hex("#AA0000")
                | yuio.color.Color.back_from_hex("#0000AA"),
                yuio.color.Color.fore_from_hex("#00AA00")
                | yuio.color.Color.back_from_hex("#00AA00"),
            ],
            [0, 0.5, 1],
            [
                yuio.color.Color.fore_from_hex("#AA0000")
                | yuio.color.Color.back_from_hex("#0000AA"),
                yuio.color.Color.fore_from_hex("#555500")
                | yuio.color.Color.back_from_hex("#005555"),
                yuio.color.Color.fore_from_hex("#00AA00")
                | yuio.color.Color.back_from_hex("#00AA00"),
            ],
        ),
    ],
)
def test_lerp(colors, coeffs, expect):
    lerp = yuio.color.Color.lerp(*colors)
    result = [lerp(c) for c in coeffs]
    assert result == expect


def test_lerp_fail():
    with pytest.raises(ValueError):
        yuio.color.Color.lerp()


@pytest.mark.parametrize(
    ("color", "expect"),
    [
        (
            yuio.color.ColorValue.from_hex("#005555").lighten(0.5),
            yuio.color.ColorValue.from_hex("#00AAAA"),
        ),
        (
            yuio.color.ColorValue.from_hex("#005555").darken(0.5),
            yuio.color.ColorValue.from_hex("#002A2A"),
        ),
        (yuio.color.ColorValue("0").darken(0.5), yuio.color.ColorValue("0")),
        (
            yuio.color.ColorValue.from_hex("#005555").match_luminosity(
                yuio.color.ColorValue.from_hex("#002A2A")
            ),
            yuio.color.ColorValue.from_hex("#002A2A"),
        ),
        (
            yuio.color.ColorValue.from_hex("#005555").match_luminosity(
                yuio.color.ColorValue.from_hex("#AA4700")
            ),
            yuio.color.ColorValue.from_hex("#00AAAA"),
        ),
        (
            yuio.color.ColorValue(0).match_luminosity(
                yuio.color.ColorValue.from_hex("#002A2A")
            ),
            yuio.color.ColorValue(0),
        ),
        (
            yuio.color.ColorValue.from_hex("#005555").match_luminosity(
                yuio.color.ColorValue(0)
            ),
            yuio.color.ColorValue.from_hex("#005555"),
        ),
    ],
)
def test_color_operations(color, expect):
    assert color == expect


@pytest.mark.parametrize(
    ("color", "cap", "expect"),
    [
        (yuio.color.Color.NONE, yuio.color.ColorSupport.NONE, ""),
        (yuio.color.Color.NONE, yuio.color.ColorSupport.ANSI, "\x1b[m"),
        (yuio.color.Color.NONE, yuio.color.ColorSupport.ANSI_256, "\x1b[m"),
        (yuio.color.Color.NONE, yuio.color.ColorSupport.ANSI_TRUE, "\x1b[m"),
        (yuio.color.Color.FORE_RED, yuio.color.ColorSupport.NONE, ""),
        (yuio.color.Color.FORE_RED, yuio.color.ColorSupport.ANSI, "\x1b[;31m"),
        (yuio.color.Color.FORE_RED, yuio.color.ColorSupport.ANSI_256, "\x1b[;31m"),
        (yuio.color.Color.FORE_RED, yuio.color.ColorSupport.ANSI_TRUE, "\x1b[;31m"),
        (yuio.color.Color.BACK_RED, yuio.color.ColorSupport.NONE, ""),
        (yuio.color.Color.BACK_RED, yuio.color.ColorSupport.ANSI, "\x1b[;41m"),
        (yuio.color.Color.BACK_RED, yuio.color.ColorSupport.ANSI_256, "\x1b[;41m"),
        (yuio.color.Color.BACK_RED, yuio.color.ColorSupport.ANSI_TRUE, "\x1b[;41m"),
        (yuio.color.Color.fore_from_hex("#338F15"), yuio.color.ColorSupport.NONE, ""),
        (
            yuio.color.Color.fore_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI,
            "\x1b[;32m",
        ),
        (
            yuio.color.Color.fore_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI_256,
            "\x1b[;38;5;64m",
        ),
        (
            yuio.color.Color.fore_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI_TRUE,
            "\x1b[;38;2;51;143;21m",
        ),
        (yuio.color.Color.back_from_hex("#338F15"), yuio.color.ColorSupport.NONE, ""),
        (
            yuio.color.Color.back_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI,
            "\x1b[;42m",
        ),
        (
            yuio.color.Color.back_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI_256,
            "\x1b[;48;5;64m",
        ),
        (
            yuio.color.Color.back_from_hex("#338F15"),
            yuio.color.ColorSupport.ANSI_TRUE,
            "\x1b[;48;2;51;143;21m",
        ),
        (
            # Special case for grayscale.
            yuio.color.Color.fore_from_hex("#333333"),
            yuio.color.ColorSupport.ANSI_256,
            "\x1b[;38;5;236m",
        ),
        (yuio.color.Color.STYLE_BOLD, yuio.color.ColorSupport.ANSI, "\x1b[;1m"),
        (yuio.color.Color.STYLE_DIM, yuio.color.ColorSupport.ANSI, "\x1b[;2m"),
        (yuio.color.Color.STYLE_ITALIC, yuio.color.ColorSupport.ANSI, "\x1b[;3m"),
        (yuio.color.Color.STYLE_UNDERLINE, yuio.color.ColorSupport.ANSI, "\x1b[;4m"),
        (yuio.color.Color.STYLE_BLINK, yuio.color.ColorSupport.ANSI, "\x1b[;5m"),
        (yuio.color.Color.STYLE_INVERSE, yuio.color.ColorSupport.ANSI, "\x1b[;7m"),
        (
            yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_ITALIC
            | yuio.color.Color.STYLE_UNDERLINE
            | yuio.color.Color.STYLE_INVERSE
            | yuio.color.Color.STYLE_BLINK,
            yuio.color.ColorSupport.ANSI,
            "\x1b[;1;2;3;4;5;7m",
        ),
        (
            yuio.color.Color.STYLE_BOLD
            | yuio.color.Color.STYLE_DIM
            | yuio.color.Color.STYLE_ITALIC
            | yuio.color.Color.STYLE_UNDERLINE
            | yuio.color.Color.STYLE_INVERSE
            | yuio.color.Color.STYLE_BLINK
            | yuio.color.Color.STYLE_NORMAL,
            yuio.color.ColorSupport.ANSI,
            "\x1b[m",
        ),
        (yuio.color.Color.STYLE_NORMAL, yuio.color.ColorSupport.ANSI, "\x1b[m"),
        (yuio.color.Color.FORE_NORMAL_DIM, yuio.color.ColorSupport.ANSI, "\x1b[;2m"),
    ],
)
def test_color_to_code(color: yuio.color.Color, cap, expect):
    assert color.as_code(cap) == expect



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

from __future__ import annotations

import yuio.app
import yuio.io
import yuio.term
import yuio.widget


@yuio.app.app
def main(
    #: When enabled, prints raw input from terminal into stderr.
    raw: bool = yuio.app.field(default=False, flags=["-r", "--raw"]),
    #: When enabled, puts terminal into bracketed paste mode.
    bracketed_paste: bool = True,
    #: When enabled, enables Kitty's key disambiguation protocol.
    modify_keyboard: bool = False,
):
    """
    Debug how Yuio parses escape sequences coming from the terminal.

    """

    # Note: we use stderr as a tty stream even though we write into stdout.
    # This is to allow redirecting stdout to some file while recording keystrokes.
    # In fact, we do exactly this in our tests.
    term = yuio.io.get_term()
    if not term.can_run_widgets:
        raise yuio.app.AppError("Stderr is not interactive")

    yuio.io.info(
        "Printing everything that the terminal sends us. Press `Ctrl+C` to stop."
    )
    with yuio.term._enter_raw_mode(
        term.ostream,
        term.istream,
        bracketed_paste=bracketed_paste,
        modify_keyboard=modify_keyboard,
    ):
        if raw:
            _read_keycode = yuio.term._read_keycode

            def read_keycode(*a, **kw):
                keycode = _read_keycode(*a, **kw)
                yuio.io.info("<c d>Read: %r</c>", keycode, to_stdout=True)
                return keycode

            yuio.term._read_keycode = read_keycode
        for event in yuio.widget._event_stream(term.ostream, term.istream):
            if isinstance(event.key, str):
                r = repr(event.key)
            else:
                r = event.key.name
            if event.alt:
                r = f"Alt+{r}"
            if event.ctrl:
                r = f"Ctrl+{r}"
            if event.shift:
                r = f"Shift+{r}"
            if event.paste_str:
                p = f" {event.paste_str!r}"
            else:
                p = ""
            yuio.io.info("Key: <c b>%s</c>`%s`", r, p, to_stdout=True)
            if event == yuio.widget.KeyboardEvent("c", ctrl=True):
                raise KeyboardInterrupt()


if __name__ == "__main__":
    main.run()



import yuio.app
import yuio.io

@yuio.app.app
def main():
    yuio.io.hl("""
        {
            "version": "1.0.0",
            "pre-release": false,
            "post-release": false,
        }
    """, syntax="json")

if __name__ == "__main__":
    main.run()



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Yuio's primary format for higher-level io is Markdown (well, a reasonably rich subset
of it).


Formatting markdown
-------------------

.. autoclass:: MdFormatter
   :members:


.. _highlighting-code:

Highlighting code
-----------------

Yuio supports basic code highlighting; it is just enough to format help messages
for CLI, and color tracebacks when an error occurs.

.. autoclass:: SyntaxHighlighter
   :members:


Markdown AST
------------

.. warning::

   This is experimental API which can change within a minor release.

.. autofunction:: parse

.. autoclass:: AstBase
   :members:

.. autoclass:: Text
   :members:

.. autoclass:: Container
   :members:

.. autoclass:: Document
   :members:

.. autoclass:: ThematicBreak
   :members:

.. autoclass:: Heading
   :members:

.. autoclass:: Paragraph
   :members:

.. autoclass:: Quote
   :members:

.. autoclass:: Code
   :members:

.. autoclass:: ListItem
   :members:

.. autoclass:: List
   :members:


"""

from __future__ import annotations

import abc
import contextlib
import dataclasses
import math
import os
import re
from dataclasses import dataclass

import yuio.color
import yuio.string
import yuio.theme
from yuio.util import dedent as _dedent

import yuio._typing_ext as _tx
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

__all__ = [
    "AstBase",
    "AstBase",
    "Code",
    "Code",
    "Container",
    "Container",
    "Document",
    "Document",
    "Heading",
    "Heading",
    "List",
    "List",
    "ListItem",
    "ListItem",
    "MdFormatter",
    "Paragraph",
    "Paragraph",
    "Quote",
    "Quote",
    "Raw",
    "SyntaxHighlighter",
    "Text",
    "Text",
    "ThematicBreak",
    "ThematicBreak",
    "parse",
]

T = _t.TypeVar("T")
TAst = _t.TypeVar("TAst", bound="AstBase")


@_t.final
class MdFormatter:
    """
    A simple markdown formatter suitable for displaying rich text in the terminal.

    :param ctx:
        a :class:`~yuio.string.ReprContext` that's used to colorize or wrap
        rendered markdown.
    :param allow_headings:
        if set to :data:`False`, headings are rendered as paragraphs.

    All CommonMark block markup except tables is supported:

    - headings:

      .. code-block:: markdown

         # Heading 1
         ## Heading 2

      Yuio has only two levels of headings. Headings past level two will look the same
      as level two headings (you can adjust theme to change this).

      If `allow_headings` is set to :data:`False`, headings look like paragraphs.

    - lists, numbered lists, quotes:

      .. code-block:: markdown

         -  List item 1,
         -  list item 2.

         1. Numbered list item 1,
         1. numbered list item 2.

         > Quoted text.

    - fenced code blocks with minimal syntax highlighting
      (see :class:`SyntaxHighlighter`):

      .. code-block:: markdown

         ```python
         for i in range(5, 8):
             print(f"Hello, world! This is {{i}}th day past the apocalypse.")
         ```

      Yuio supports ``python``, ``traceback``, ``bash``, ``diff``,
      and ``json`` syntaxes.

    Inline markdown only handles inline code blocks:

    .. code-block:: markdown

       This is `code`. It will be rendered as code.
       Other inline styles, such as _italic_, are not supported!

    However, color tags are supported, so you can highlight text as follows:

    .. code-block:: markdown

       This is <c b>bold text</c>. It will be rendered bold.

    """

    def __init__(
        self,
        ctx: yuio.string.ReprContext,
        *,
        allow_headings: bool = True,
    ):
        self._ctx = ctx
        self.allow_headings: bool = allow_headings

        self._is_first_line: bool
        self._out: list[yuio.string.ColorizedString]
        self._indent: yuio.string.ColorizedString
        self._continuation_indent: yuio.string.ColorizedString

    @property
    def ctx(self):
        return self._ctx

    @property
    def width(self):
        return self._ctx.width

    def format(
        self, md: str, *, dedent: bool = True
    ) -> list[yuio.string.ColorizedString]:
        """
        Format a markdown document.

        :param md:
            markdown to format. Common indentation will be removed from this string,
            making it suitable to use with triple quote literals.
        :param dedent:
            remove lading indent from markdown.
        :returns:
            rendered markdown as a list of individual lines without newline
            characters at the end.

        """

        return self.format_node(self.parse(md, dedent=dedent))

    def parse(self, md: str, /, *, dedent: bool = True) -> Document:
        """
        Parse a markdown document and return an AST node.

        .. warning::

           This is experimental API which can change within a minor release.

        :param md:
            markdown to parse. Common indentation will be removed from this string,
            making it suitable to use with triple quote literals.
        :param dedent:
            remove lading indent from markdown.
        :returns:
            parsed AST node.

        """

        if dedent:
            md = _dedent(md)

        return _MdParser(self.allow_headings).parse(md)

    def format_node(self, node: AstBase, /) -> list[yuio.string.ColorizedString]:
        """
        Format a parsed markdown document.

        .. warning::

           This is an experimental API which can change within a minor release.

        :param md:
            AST node to format.
        :returns:
            rendered markdown as a list of individual lines without newline
            characters at the end.

        """

        self._is_first_line = True
        self._separate_paragraphs = True
        self._out = []
        self._indent = yuio.string.ColorizedString()
        self._continuation_indent = yuio.string.ColorizedString()

        self._format(node)

        return self._out

    def colorize(
        self,
        text: str,
        /,
        *,
        default_color: yuio.color.Color | str = yuio.color.Color.NONE,
    ):
        """
        Parse and colorize contents of a paragraph.

        This is a shortcut for calling :func:`colorize` with this formatter's theme.

        :param line:
            text to colorize.
        :param default_color:
            color or color tag to apply to the entire text.
        :returns:
            a colorized string.

        """

        return yuio.string.colorize(text, default_color=default_color, ctx=self.ctx)

    @contextlib.contextmanager
    def _with_indent(
        self,
        color: yuio.color.Color | str | None,
        s: yuio.string.AnyString,
        /,
        *,
        continue_with_spaces: bool = True,
    ):
        color = self.ctx.to_color(color)
        indent = yuio.string.ColorizedString(color)
        indent += s

        old_indent = self._indent
        old_continuation_indent = self._continuation_indent

        if continue_with_spaces:
            continuation_indent = yuio.string.ColorizedString(" " * indent.width)
        else:
            continuation_indent = indent

        self._indent = self._indent + indent
        self._continuation_indent = self._continuation_indent + continuation_indent

        try:
            yield
        finally:
            self._indent = old_indent
            self._continuation_indent = old_continuation_indent

    def _line(self, line: yuio.string.ColorizedString, /):
        self._out.append(line)

        self._is_first_line = False
        self._indent = self._continuation_indent

    def _format(self, node: AstBase, /):
        getattr(self, f"_format_{node.__class__.__name__.lstrip('_')}")(node)

    def _format_Raw(self, node: Raw, /):
        for line in node.raw.wrap(
            self.width,
            indent=self._indent,
            continuation_indent=self._continuation_indent,
            break_long_nowrap_words=True,
        ):
            self._line(line)

    def _format_Text(self, node: Text, /, *, default_color: yuio.color.Color):
        s = self.colorize(
            "\n".join(node.lines).strip(),
            default_color=default_color,
        )

        for line in s.wrap(
            self.width,
            indent=self._indent,
            continuation_indent=self._continuation_indent,
            preserve_newlines=False,
            break_long_nowrap_words=True,
        ):
            self._line(line)

    def _format_Container(self, node: Container[TAst], /):
        self._is_first_line = True
        for item in node.items:
            if not self._is_first_line and self._separate_paragraphs:
                self._line(self._indent)
            self._format(item)

    def _format_Document(self, node: Document, /):
        self._format_Container(node)

    def _format_ThematicBreak(self, _: ThematicBreak):
        decoration = self.ctx.get_msg_decoration("thematic_break")
        self._line(self._indent + decoration)

    def _format_Heading(self, node: Heading, /):
        if not self._is_first_line:
            self._line(self._indent)

        decoration = self.ctx.get_msg_decoration(f"heading/{node.level}")
        with self._with_indent(f"msg/decoration:heading/{node.level}", decoration):
            self._format_Text(
                node,
                default_color=self.ctx.get_color(f"msg/text:heading/{node.level}"),
            )

        self._line(self._indent)
        self._is_first_line = True

    def _format_Paragraph(self, node: Paragraph, /):
        self._format_Text(node, default_color=self.ctx.get_color("msg/text:paragraph"))

    def _format_ListItem(self, node: ListItem, /, *, min_width: int = 0):
        decoration = self.ctx.get_msg_decoration("list")
        if node.number is not None:
            decoration = f"{node.number:>{min_width}}." + " " * (
                yuio.string.line_width(decoration) - min_width - 1
            )
        with self._with_indent("msg/decoration:list", decoration):
            self._format_Container(node)

    def _format_Quote(self, node: Quote, /):
        decoration = self.ctx.get_msg_decoration("quote")
        with self._with_indent(
            "msg/decoration:quote", decoration, continue_with_spaces=False
        ):
            self._format_Container(node)

    def _format_Code(self, node: Code, /):
        s = SyntaxHighlighter.get_highlighter(node.syntax).highlight(
            self.ctx.theme,
            "\n".join(node.lines),
        )

        decoration = self.ctx.get_msg_decoration("code")
        with self._with_indent("msg/decoration:code", decoration):
            self._line(
                s.indent(
                    indent=self._indent,
                    continuation_indent=self._continuation_indent,
                )
            )

    def _format_List(self, node: List, /):
        max_number = max(item.number or 0 for item in node.items)
        min_width = math.ceil(math.log10(max_number)) if max_number > 0 else 1
        self._is_first_line = True
        for item in node.items:
            if not self._is_first_line:
                self._line(self._indent)
            self._format_ListItem(item, min_width=min_width)


@dataclass(kw_only=True, slots=True)
class AstBase(abc.ABC):
    """
    Base class for all AST nodes that represent parsed markdown document.

    """

    def _dump_params(self) -> str:
        s = self.__class__.__name__.lstrip("_")
        for field in dataclasses.fields(self):
            if field.repr:
                s += f" {getattr(self, field.name)!r}"
        return s

    def dump(self, indent: str = "") -> str:
        """
        Dump an AST node into a lisp-like text representation.

        """

        return f"{indent}({self._dump_params()})"


@dataclass(kw_only=True, slots=True)
class Raw(AstBase):
    """
    Embeds already formatted paragraph into the document.

    """

    raw: yuio.string.ColorizedString
    """
    Raw colorized string to add to the document.

    """


@dataclass(kw_only=True, slots=True)
class Text(AstBase):
    """
    Base class for all text-based AST nodes, i.e. paragraphs, headings, etc.

    """

    lines: list[str] = dataclasses.field(repr=False)
    """
    Text lines as parsed from the original document.

    """

    def dump(self, indent: str = "") -> str:
        s = f"{indent}({self._dump_params()}"
        indent += "  "
        for line in self.lines:
            s += "\n" + indent
            s += repr(line)
        s += ")"
        return s


@dataclass(kw_only=True, slots=True)
class Container(AstBase, _t.Generic[TAst]):
    """
    Base class for all container-based AST nodes, i.e. list items or quotes.

    This class works as a list of items. Usually it contains arbitrary AST nodes,
    but it can also be limited to specific kinds of nodes via its generic variable.

    """

    items: list[TAst] = dataclasses.field(repr=False)
    """
    Inner AST nodes in the container.

    """

    def dump(self, indent: str = "") -> str:
        s = f"{indent or ''}({self._dump_params()}"
        indent += "  "
        for items in self.items:
            s += "\n"
            s += items.dump(indent)
        s += ")"
        return s


@dataclass(kw_only=True, slots=True)
class Document(Container[AstBase]):
    """
    Root node that contains the entire markdown document.

    """


@dataclass(kw_only=True, slots=True)
class ThematicBreak(AstBase):
    """
    Represents a visual break in text, a.k.a. an asterism.

    """


@dataclass(kw_only=True, slots=True)
class Heading(Text):
    """
    Represents a heading.

    """

    level: int
    """
    Level of the heading, `1`-based.

    """


@dataclass(kw_only=True, slots=True)
class Paragraph(Text):
    """
    Represents a regular paragraph.

    """


@dataclass(kw_only=True, slots=True)
class Quote(Container[AstBase]):
    """
    Represents a quotation block.

    """


@dataclass(kw_only=True, slots=True)
class Code(Text):
    """
    Represents a highlighted block of code.

    """

    syntax: str
    """
    Syntax indicator as parsed form the original document.

    """


@dataclass(kw_only=True, slots=True)
class ListItem(Container[AstBase]):
    """
    A possibly numbered element of a list.

    """

    number: int | None
    """
    If present, this is the item's number in a numbered list.

    """


@dataclass(kw_only=True, slots=True)
class List(Container[ListItem]):
    """
    A collection of list items.

    """


_HEADING_RE = re.compile(
    r"""
    ^
    \s{0,3}                     # - Initial indent.
    (?P<marker>\#{1,6})         # - Heading marker.
    (?P<text>\s.*?)?            # - Heading text. Unless empty, text must be separated
                                #   from the heading marker by a space.
    (?:(?<=\s)\#+)?             # - Optional closing hashes. Must be separated from
                                #   the previous content by a space. We use lookbehind
                                #   here, because if the text is empty, the space
                                #   between heading marker and closing hashes will be
                                #   matched by the `text` group.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_SETEXT_HEADING_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<level>-|=)              # - Heading underline.
    \2*                         # - More heading underline.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_LIST_RE = re.compile(
    r"""
    ^
    (?P<marker>
      \s{0,3}                   # - Initial indent.
      (?P<type>[-*+])           # - List marker.
      (?:
          \s(?:\s{0,3}(?=\S))?  # - One mandatory and up to three optional spaces;
                                #   When there are more than three optional spaces,
                                #   we treat then as a list marker followed
                                #   by a single space, followed by a code block.
        | $))                   # - For cases when a list starts with an empty line.
    (?P<text>.*)                # - Text of the first line in the list.
    $
    """,
    re.VERBOSE,
)
_NUMBERED_LIST_RE = re.compile(
    r"""
    ^
    (?P<marker>
      \s{0,3}                   # - Initial indent.
      (?P<number>\d{1,9})       # - Number.
      (?P<type>[.:)])           # - Numbered list marker.
      (?:
          \s(?:\s{0,3}(?=\S))?  # - One mandatory and up to three optional spaces;
                                #   When there are more than three optional spaces,
                                #   we treat then as a list marker followed
                                #   by a single space, followed by a code block.
        | $))                   # - For cases when a list starts with an empty line.
    (?P<text>.*)                # - Text of the first line in the list.
    $
    """,
    re.VERBOSE,
)
_CODE_BACKTICK_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>```+)             # - Backtick fence.
    (?P<syntax>[^`]*)           # - Syntax, can't contain backtick.
    $
    """,
    re.VERBOSE,
)
_CODE_TILDE_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>~~~+)             # - Backtick fence.
    (?P<syntax>.*)              # - Syntax, can be anything.
    $
    """,
    re.VERBOSE,
)
_CODE_FENCE_END_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>~~~+|```+)        # - Fence.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_CODE_RE = re.compile(
    r"""
    ^
    \s{4}                       # - Initial code indent.
    (?P<text>.*)                # - First code line.
    $
    """,
    re.VERBOSE,
)
_QUOTE_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial quote indent.
    >                           # - Quote marker.
    \s?                         # - Optional space after the marker.
    (?P<text>.*)                # - Text of the first line in the quote.
    $
    """,
    re.VERBOSE,
)
_THEMATIC_BREAK_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial quote indent.
    ([-*_])\s*(\2\s*){2,}       # - At least three break characters separated by spaces.
    $
    """,
    re.VERBOSE,
)
_LINE_FEED_RE = re.compile(r"\r\n|\r|\n")


class _MdParser:
    @dataclass(kw_only=True, slots=True)
    class Default:
        pass

    @dataclass(kw_only=True, slots=True)
    class List:
        type: str
        marker_len: int
        list: List
        parser: _MdParser
        number: int | None = None

    @dataclass(kw_only=True, slots=True)
    class Quote:
        parser: _MdParser

    @dataclass(kw_only=True, slots=True)
    class Code:
        lines: list[str]

    @dataclass(kw_only=True, slots=True)
    class FencedCode:
        indent: int
        fence_symbol: str
        fence_length: int
        syntax: str
        lines: list[str]

    @dataclass(kw_only=True, slots=True)
    class Paragraph:
        lines: list[str]

    State: _t.TypeAlias = Default | List | Quote | Code | FencedCode | Paragraph

    def __init__(self, allow_headings: bool = True):
        self._allow_headings = allow_headings
        self._nodes: list[AstBase] = []
        self._state: _MdParser.State = self.Default()

    def _parser(self) -> _MdParser:
        return _MdParser(self._allow_headings)

    @staticmethod
    def _is_blank(s: str) -> bool:
        return not s or s.isspace()

    def parse(self, s: str) -> Document:
        s = s.expandtabs(tabsize=4)
        for line in _LINE_FEED_RE.split(s):
            self._handle_line(line)
        return Document(items=self._finalize())

    def _handle_line(self, line: str):
        getattr(self, f"_handle_line_{self._state.__class__.__name__}")(line)

    def _handle_lazy_line(self, line: str) -> bool:
        return getattr(self, f"_handle_lazy_line_{self._state.__class__.__name__}")(
            line
        )

    def _flush(self):
        getattr(self, f"_flush_{self._state.__class__.__name__}")()

    def _handle_line_List(self, line: str):
        assert type(self._state) is self.List
        if not line or line[: self._state.marker_len].isspace():
            self._state.parser._handle_line(line[self._state.marker_len :])
        elif (
            (match := _LIST_RE.match(line)) or (match := _NUMBERED_LIST_RE.match(line))
        ) and match.group("type") == self._state.type:
            item = ListItem(
                items=self._state.parser._finalize(),
                number=self._state.number,
            )
            self._state.list.items.append(item)
            self._state.marker_len = len(match.group("marker"))
            self._state.parser._handle_line(match.group("text"))
            if self._state.number is not None:
                self._state.number += 1
        elif not self._state.parser._handle_lazy_line(line):
            self._flush_List()
            self._handle_line_Default(line)

    def _handle_lazy_line_List(self, line: str) -> bool:
        assert type(self._state) is self.List
        if self._state.parser._handle_lazy_line(line):
            return True
        return False

    def _flush_List(self):
        assert type(self._state) is self.List
        item = ListItem(
            items=self._state.parser._finalize(),
            number=self._state.number,
        )
        self._state.list.items.append(item)
        self._nodes.append(self._state.list)
        self._state = self.Default()

    def _handle_line_Quote(self, line: str):
        assert type(self._state) is self.Quote
        if match := _QUOTE_RE.match(line):
            self._state.parser._handle_line(match.group("text"))
        elif self._is_blank(line) or not self._state.parser._handle_lazy_line(line):
            self._flush_Quote()
            self._handle_line_Default(line)

    def _handle_lazy_line_Quote(self, line: str) -> bool:
        assert type(self._state) is self.Quote
        if self._state.parser._handle_lazy_line(line):
            return True
        else:
            return False

    def _flush_Quote(self):
        assert type(self._state) is self.Quote
        self._nodes.append(Quote(items=self._state.parser._finalize()))
        self._state = self.Default()

    def _handle_line_Code(self, line: str):
        assert type(self._state) is self.Code
        if self._is_blank(line) or line.startswith("    "):
            self._state.lines.append(line[4:])
        else:
            self._flush_Code()
            self._handle_line_Default(line)

    def _handle_lazy_line_Code(self, line: str) -> bool:
        assert type(self._state) is self.Code
        return False  # No lazy continuations for code!

    def _flush_Code(self):
        assert type(self._state) is self.Code
        while self._state.lines and self._is_blank(self._state.lines[-1]):
            self._state.lines.pop()
        self._nodes.append(
            Code(
                lines=self._state.lines,
                syntax="",
            )
        )
        self._state = self.Default()

    def _handle_line_FencedCode(self, line: str):
        assert type(self._state) is self.FencedCode
        if (
            (match := _CODE_FENCE_END_RE.match(line))
            and match.group("fence")[0] == self._state.fence_symbol
            and len(match.group("fence")) == self._state.fence_length
        ):
            self._flush_FencedCode()
        else:
            if self._state.indent == 0:
                pass
            elif line[: self._state.indent].isspace():
                line = line[self._state.indent :]
            else:
                line = line.lstrip()
            self._state.lines.append(line)

    def _handle_lazy_line_FencedCode(self, line: str) -> bool:
        assert type(self._state) is self.FencedCode
        return False

    def _flush_FencedCode(self):
        assert type(self._state) is self.FencedCode
        self._nodes.append(
            Code(
                lines=self._state.lines,
                syntax=self._state.syntax,
            )
        )
        self._state = self.Default()

    def _handle_line_Paragraph(self, line: str):
        assert type(self._state) is self.Paragraph
        if match := _SETEXT_HEADING_RE.match(line):
            level = 1 if match.group("level") == "=" else 2
            self._nodes.append(
                Heading(
                    lines=self._state.lines,
                    level=level,
                )
            )
            self._state = self.Default()
        elif (
            self._is_blank(line)
            or _THEMATIC_BREAK_RE.match(line)
            or (self._allow_headings and _HEADING_RE.match(line))
            or _CODE_BACKTICK_RE.match(line)
            or _CODE_TILDE_RE.match(line)
            or _LIST_RE.match(line)
            or _NUMBERED_LIST_RE.match(line)
            or _QUOTE_RE.match(line)
        ):
            self._flush_Paragraph()
            self._handle_line_Default(line)
        else:
            self._state.lines.append(line)

    def _handle_lazy_line_Paragraph(self, line: str) -> bool:
        assert type(self._state) is self.Paragraph
        if (
            self._is_blank(line)
            or _THEMATIC_BREAK_RE.match(line)
            or (self._allow_headings and _HEADING_RE.match(line))
            or _CODE_BACKTICK_RE.match(line)
            or _CODE_TILDE_RE.match(line)
            or _LIST_RE.match(line)
            or _NUMBERED_LIST_RE.match(line)
            or _QUOTE_RE.match(line)
        ):
            self._flush_Paragraph()
            return False
        else:
            self._state.lines.append(line)
            return True

    def _flush_Paragraph(self):
        assert type(self._state) is self.Paragraph
        self._nodes.append(Paragraph(lines=self._state.lines))
        self._state = self.Default()

    def _handle_line_Default(self, line: str):
        assert type(self._state) is self.Default
        if self._is_blank(line):
            pass  # do nothing
        elif _THEMATIC_BREAK_RE.match(line):
            self._nodes.append(ThematicBreak())
        elif self._allow_headings and (match := _HEADING_RE.match(line)):
            level = len(match.group("marker"))
            self._nodes.append(
                Heading(
                    lines=[match.group("text").strip()],
                    level=level,
                )
            )
        elif (match := _CODE_BACKTICK_RE.match(line)) or (
            match := _CODE_TILDE_RE.match(line)
        ):
            indent = len(match.group("indent"))
            syntax = match.group("syntax").strip()
            fence_symbol = match.group("fence")[0]
            fence_length = len(match.group("fence"))
            self._state = self.FencedCode(
                indent=indent,
                fence_symbol=fence_symbol,
                fence_length=fence_length,
                syntax=syntax,
                lines=[],
            )
        elif match := _CODE_RE.match(line):
            self._state = self.Code(lines=[match.group("text")])
        elif (match := _LIST_RE.match(line)) or (
            match := _NUMBERED_LIST_RE.match(line)
        ):
            indent = len(match.group("marker"))
            list_type = match.group("type")
            number_str = match.groupdict().get("number", None)
            number = int(number_str) if number_str else None
            self._state = self.List(
                type=list_type,
                marker_len=indent,
                list=List(items=[]),
                parser=self._parser(),
                number=number,
            )
            self._state.parser._handle_line(match.group("text"))
        elif match := _QUOTE_RE.match(line):
            self._state = self.Quote(parser=self._parser())
            self._state.parser._handle_line(match.group("text"))
        else:
            self._state = self.Paragraph(lines=[line])

    def _handle_lazy_line_Default(self, line: str) -> bool:
        assert type(self._state) is self.Default
        return False

    def _flush_Default(self):
        assert type(self._state) is self.Default

    def _finalize(self) -> list[AstBase]:
        self._flush()
        result = self._nodes
        self._nodes = []
        return result


def parse(md: str, /, *, dedent: bool = True, allow_headings: bool = True) -> Document:
    """
    Parse a markdown document and return an AST node.

    :param md:
        markdown to parse. Common indentation will be removed from this string,
        making it suitable to use with triple quote literals.
    :param dedent:
        remove lading indent from markdown.
    :param allow_headings:
        if set to :data:`False`, headings are rendered as paragraphs.
    :returns:
        parsed AST node.

    """

    if dedent:
        md = _dedent(md)

    return _MdParser(allow_headings).parse(md)


_SYNTAXES: dict[str, SyntaxHighlighter] = {}
"""
Global syntax registry.

"""


class SyntaxHighlighter(abc.ABC):
    @property
    @abc.abstractmethod
    def syntaxes(self) -> list[str]:
        """
        List of syntax names that should be associated with this highlighter.

        """

        return []

    @property
    def syntax(self) -> str:
        """
        The primary syntax name for this highlighter, defaults to the first element
        of the :attr:`~SyntaxHighlighter.syntaxes` list.

        This name is used to look up colors in a theme.

        """

        return self.syntaxes[0] if self.syntaxes else "unknown"

    @classmethod
    def register_highlighter(cls, highlighter: SyntaxHighlighter):
        """
        Register a highlighter in a global registry, and allow looking it up
        via the :meth:`~SyntaxHighlighter.get_highlighter` method.

        :param highlighter:
            a highlighter instance.

        """

        for syntax in highlighter.syntaxes:
            _SYNTAXES[syntax.lower().replace("_", "-")] = highlighter

    @classmethod
    def get_highlighter(cls, syntax: str, /) -> SyntaxHighlighter:
        """
        Look up highlighter by a syntax name.

        :param syntax:
            name of the syntax highlighter.
        :returns:
            a highlighter instance.

            If highlighter with the given name can't be found, returns a dummy
            highlighter that does nothing.

        """

        return _SYNTAXES.get(
            syntax.lower().replace("_", "-"),
            _DummySyntaxHighlighter(),
        )

    @abc.abstractmethod
    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.color.Color | str | None = None,
    ) -> yuio.string.ColorizedString:
        """
        Highlight the given code using the given theme.

        :param theme:
            theme that will be used to look up color tags.
        :param code:
            code to highlight.
        :param default_color:
            color or color tag to apply to the entire code.

        """

        raise NotImplementedError()

    def _get_default_color(
        self,
        theme: yuio.theme.Theme,
        default_color: yuio.color.Color | str | None,
    ) -> yuio.color.Color:
        return theme.to_color(default_color) | theme.get_color(
            f"msg/text:code/{self.syntax}"
        )


class _DummySyntaxHighlighter(SyntaxHighlighter):
    @property
    def syntaxes(self) -> list[str]:
        return ["text", "plain-text"]

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.color.Color | str | None = None,
    ) -> yuio.string.ColorizedString:
        return yuio.string.ColorizedString(
            [
                self._get_default_color(theme, default_color),
                code,
                yuio.color.Color.NONE,
            ]
        )


SyntaxHighlighter.register_highlighter(_DummySyntaxHighlighter())


class _ReSyntaxHighlighter(SyntaxHighlighter):
    def __init__(
        self,
        syntaxes: list[str],
        pattern: _tx.StrRePattern,
        str_esc_pattern: _tx.StrRePattern | None = None,
    ):
        self._syntaxes = syntaxes
        self._pattern = pattern
        self._str_esc_pattern = str_esc_pattern

    @property
    def syntaxes(self) -> list[str]:
        return self._syntaxes

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.color.Color | str | None = None,
    ) -> yuio.string.ColorizedString:
        default_color = self._get_default_color(theme, default_color)

        raw = yuio.string.ColorizedString()

        last_pos = 0
        for code_unit in self._pattern.finditer(code):
            if last_pos < code_unit.start():
                raw += default_color
                raw += code[last_pos : code_unit.start()]
            last_pos = code_unit.end()

            for name, text in sorted(code_unit.groupdict().items()):
                if not text:
                    continue
                name = name.split("__", maxsplit=1)[-1]
                if self._str_esc_pattern is not None and name == "str":
                    str_color = default_color | theme.get_color(f"hl/str:{self.syntax}")
                    esc_color = default_color | theme.get_color(
                        f"hl/str/esc:{self.syntax}"
                    )
                    last_escape_pos = 0
                    for escape_unit in self._str_esc_pattern.finditer(text):
                        if last_escape_pos < escape_unit.start():
                            raw += str_color
                            raw += text[last_escape_pos : escape_unit.start()]
                        last_escape_pos = escape_unit.end()
                        if escape := text[escape_unit.start() : escape_unit.end()]:
                            raw += esc_color
                            raw += escape
                    if last_escape_pos < len(text):
                        raw += str_color
                        raw += text[last_escape_pos:]
                else:
                    raw += default_color | theme.get_color(f"hl/{name}:{self.syntax}")
                    raw += text

        if last_pos < len(code):
            raw += default_color
            raw += code[last_pos:]

        return raw


_PY_SYNTAX = re.compile(
    r"""
        (?P<kwd>
            \b(?:                                   # keyword
                and|as|assert|async|await|break|class|continue|def|del|elif|else|
                except|finally|for|from|global|if|import|in|is|lambda|
                nonlocal|not|or|pass|raise|return|try|while|with|yield
            )\b)
        | (?P<str>
            [rfut]*(                                # string prefix
                  '(?:\\.|[^\\'])*(?:'|\n)          # singly-quoted string
                | "(?:\\.|[^\\"])*(?:"|\n)          # doubly-quoted string
                | \"""(\\.|[^\\]|\n)*?\"""          # long singly-quoted string
                | '''(\\.|[^\\]|\n)*?'''))          # long doubly-quoted string
        | (?P<lit>
            (?<![\.\w])(
                  [+-]?\d+(?:\.\d*(?:e[+-]?\d+)?)?  # int or float
                | [+-]?\.\d+(?:e[+-]?\d+)?          # float that starts with dot
                | [+-]?0x[0-9a-fA-F]+               # hex
                | [+-]?0b[01]+                      # bin
                | \b(?:None|True|False)\b))         # bool or none
        | (?P<type>
            \b(?:                                   # type
                str|int|float|complex|list|tuple|range|dict|set|frozenset|bool|
                bytes|bytearray|memoryview|(?:[A-Z](?:[A-Z0-9_]*?[a-z]\w*)?)
            )\b)
        | (?P<punct>[{}()\[\]\\;,])                 # punctuation
        | (?P<comment>\#.*$)                        # comment
    """,
    re.MULTILINE | re.VERBOSE,
)
_PY_ESC_PATTERN = re.compile(
    r"""
        \\(
            \n                                      # escaped newline
            | [\\'"abfnrtv]                         # normal escape
            | [0-7]{3}                              # octal escape
            | x[0-9a-fA-F]{2}                       # hex escape
            | u[0-9a-fA-F]{4}                       # short unicode escape
            | U[0-9a-fA-F]{8}                       # long unicode escape
            | N\{[^}\n]+\}                          # unicode character names
            | [{}]                                  # template
            | %                                     # percent formatting
              (?:\([^)]*\))?                        # mapping key
              [#0\-+ ]*                             # conversion Flag
              (?:\*|\d+)?                           # field width
              (?:\.(?:\*|\d*))?                     # precision
              [hlL]?                                # unused length modifier
              .                                     # conversion type
        )
    """,
    re.VERBOSE,
)


SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["py", "py3", "py-3", "python", "python3", "python-3"],
        _PY_SYNTAX,
        str_esc_pattern=_PY_ESC_PATTERN,
    )
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["repr"],
        _PY_SYNTAX,
        str_esc_pattern=_PY_ESC_PATTERN,
    )
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["sh", "bash"],
        re.compile(
            r"""
                (?P<kwd>
                    \b(?:                                   # keyword
                      if|then|elif|else|fi|time|for|in|until|while|do|done|case|
                      esac|coproc|select|function
                    )\b
                  | \[\[                                    # `test` syntax: if [[ ... ]]
                  | \]\])
                | (?P<a0__punct>(?:^|\|\|?|&&|\$\())        # chaining operator: pipe or logic
                  (?P<a1__>\s*)
                  (?P<a2__prog>[\w./~]([\w.@/-]|\\.)+)      # prog
                | (?P<str>
                      '[^']*'                               # singly-quoted string
                    | "(?:\\.|[^\\"])*")                    # doubly-quoted string
                | (?P<punct>
                      [{}()\[\]\\;!&|]                      # punctuation
                    | <{1,3}                                # input redirect
                    | [12]?>{1,2}(?:&[12])?)                # output redirect
                | (?P<comment>\#.*$)                        # comment
                | (?P<flag>(?<![\w-])-[a-zA-Z0-9_-]+\b)     # flag
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    ),
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["sh-usage", "bash-usage"],
        re.compile(
            r"""
                (?P<kwd>
                  \b(?:                                     # keyword
                    if|then|elif|else|fi|time|for|in|until|while|do|done|case|
                    esac|coproc|select|function
                  )\b)
                | (?P<prog>%\(prog\)s)                      # prog
                | (?P<str>
                      '[^']*'                               # singly-quoted string
                    | "(?:\\.|[^\\"])*")                    # doubly-quoted string
                | (?P<comment>\#.*$)                        # comment
                | (?P<flag>(?<![\w-])
                      -[-\w]+\b                             # flag
                    | <options>                             # options
                  )
                | (?P<metavar><[^>]+>)                      # metavar
                | (?P<punct>[{}()\[\]\\;!&|])               # punctuation
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    )
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["diff"],
        re.compile(
            r"""
                (?P<meta>^(\-\-\-|\+\+\+|\@\@)[^\r\n]*$)
                | (?P<added>^\+[^\r\n]*$)
                | (?P<removed>^\-[^\r\n]*$)
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    ),
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["json"],
        re.compile(
            r"""
                (?P<lit>\b(?:true|false|null)\b)            # keyword
                | (?P<str>"(?:\\.|[^\\"])*(?:"|\n))         # doubly-quoted string
                | (?P<punct>[{}\[\],:])                     # punctuation
            """,
            re.MULTILINE | re.VERBOSE,
        ),
        str_esc_pattern=re.compile(
            r"""
                \\(
                    \n
                    | [\\/"bfnrt]
                    | u[0-9a-fA-F]{4}
                )
            """,
            re.VERBOSE,
        ),
    ),
)


class _TbHighlighter(SyntaxHighlighter):
    @property
    def syntaxes(self) -> list[str]:
        return [
            "tb",
            "traceback",
            "py-tb",
            "py3-tb",
            "py-3-tb",
            "py-traceback",
            "py3-traceback",
            "py-3-traceback",
            "python-tb",
            "python3-tb",
            "python-3-tb",
            "python-traceback",
            "python3-traceback",
            "python-3-traceback",
        ]

    class _StackColors:
        def __init__(
            self, theme: yuio.theme.Theme, default_color: yuio.color.Color, tag: str
        ):
            self.file_color = default_color | theme.get_color(f"tb/frame/{tag}/file")
            self.file_path_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/path"
            )
            self.file_line_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/line"
            )
            self.file_module_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/module"
            )
            self.code_color = default_color | theme.get_color(f"tb/frame/{tag}/code")
            self.highlight_color = default_color | theme.get_color(
                f"tb/frame/{tag}/highlight"
            )

    _TB_RE = re.compile(
        r"^(?P<indent>[ |+]*)(Stack|Traceback|Exception Group Traceback) \(most recent call last\):$"
    )
    _TB_MSG_RE = re.compile(r"^(?P<indent>[ |+]*)[A-Za-z_][A-Za-z0-9_]*($|:.*$)")
    _TB_LINE_FILE = re.compile(
        r'^[ |+]*File (?P<file>"[^"]*"), line (?P<line>\d+)(?:, in (?P<loc>.*))?$'
    )
    _TB_LINE_HIGHLIGHT = re.compile(r"^[ |+^~-]*$")
    _SITE_PACKAGES = os.sep + "lib" + os.sep + "site-packages" + os.sep
    _LIB_PYTHON = os.sep + "lib" + os.sep + "python"

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.color.Color | str | None = None,
    ) -> yuio.string.ColorizedString:
        default_color = self._get_default_color(theme, default_color)

        py_highlighter = SyntaxHighlighter.get_highlighter("python")

        heading_color = default_color | theme.get_color("tb/heading")
        message_color = default_color | theme.get_color("tb/message")

        stack_normal_colors = self._StackColors(theme, default_color, "usr")
        stack_lib_colors = self._StackColors(theme, default_color, "lib")
        stack_colors = stack_normal_colors

        res = yuio.string.ColorizedString()

        PLAIN_TEXT, STACK, MESSAGE = 1, 2, 3
        state = PLAIN_TEXT
        stack_indent = ""
        message_indent = ""

        for line in code.splitlines(keepends=True):
            if state is STACK:
                if line.startswith(stack_indent):
                    # We're still in the stack.
                    if match := self._TB_LINE_FILE.match(line):
                        file, line, loc = match.group("file", "line", "loc")

                        if self._SITE_PACKAGES in file or self._LIB_PYTHON in file:
                            stack_colors = stack_lib_colors
                        else:
                            stack_colors = stack_normal_colors

                        res += yuio.color.Color.NONE
                        res += stack_indent
                        res += stack_colors.file_color
                        res += "File "
                        res += stack_colors.file_path_color
                        res += file
                        res += stack_colors.file_color
                        res += ", line "
                        res += stack_colors.file_line_color
                        res += line
                        res += stack_colors.file_color

                        if loc:
                            res += ", in "
                            res += stack_colors.file_module_color
                            res += loc
                            res += stack_colors.file_color

                        res += "\n"
                    elif match := self._TB_LINE_HIGHLIGHT.match(line):
                        res += yuio.color.Color.NONE
                        res += stack_indent
                        res += stack_colors.highlight_color
                        res += line[len(stack_indent) :]
                    else:
                        res += yuio.color.Color.NONE
                        res += stack_indent
                        res += py_highlighter.highlight(
                            theme,
                            line[len(stack_indent) :],
                            stack_colors.code_color,
                        )
                    continue
                else:
                    # Stack has ended, this line is actually a message.
                    state = MESSAGE

            if state is MESSAGE:
                if line and line != "\n" and line.startswith(message_indent):
                    # We're still in the message.
                    res += yuio.color.Color.NONE
                    res += message_indent
                    res += message_color
                    res += line[len(message_indent) :]
                    continue
                else:
                    # Message has ended, this line is actually a plain text.
                    state = PLAIN_TEXT

            if state is PLAIN_TEXT:
                if match := self._TB_RE.match(line):
                    # Plain text has ended, this is actually a heading.
                    message_indent = match.group("indent").replace("+", "|")
                    stack_indent = message_indent + "  "

                    res += yuio.color.Color.NONE
                    res += message_indent
                    res += heading_color
                    res += line[len(message_indent) :]

                    state = STACK
                    continue
                elif match := self._TB_MSG_RE.match(line):
                    # Plain text has ended, this is an error message (without a traceback).
                    message_indent = match.group("indent").replace("+", "|")
                    stack_indent = message_indent + "  "

                    res += yuio.color.Color.NONE
                    res += message_indent
                    res += message_color
                    res += line[len(message_indent) :]

                    state = MESSAGE
                    continue
                else:
                    # We're still in plain text.
                    res += yuio.color.Color.NONE
                    res += line
                    continue

        return res


SyntaxHighlighter.register_highlighter(_TbHighlighter())



import time
import yuio.io
import yuio.theme
import yuio.app

class Theme(yuio.theme.DefaultTheme):
    colors = {
        "accent_color": "red",
        "accent_color_2": "blue",
    }

@yuio.app.app
def main():
    yuio.io.heading("Custom accent colors")

    with yuio.io.Task("Example task") as task:
        for i in range(100):
            task.progress(i, 100)
            time.sleep(0.05)

main.theme = Theme

if __name__ == "__main__":
    main.run()



import yuio.io
import yuio.theme
import yuio.app

class Theme(yuio.theme.DefaultTheme):
    separate_headings = False
    msg_decorations_unicode = {
        "heading/1": "=> ",
        "info": "-> ",
    }
    msg_decorations_ascii = {
        "heading/1": "=> ",
        "info": "-> ",
    }

@yuio.app.app
def main():
    yuio.io.heading("Custom decorations")
    yuio.io.info("This is an info message")

main.theme = Theme

if __name__ == "__main__":
    main.run()
